# coding=utf-8
from euphorie.client import utils
from euphorie.client.browser.webhelpers import WebHelpers
from euphorie.client.country import ClientCountry
from euphorie.client.sector import ClientSector
from euphorie.client.tests.utils import testRequest
from euphorie.client.utils import locals
from euphorie.content.survey import Survey
from euphorie.testing import EuphorieIntegrationTestCase
from OFS.SimpleItem import SimpleItem
from PIL.ImageColor import getrgb
from zope.annotation import IAnnotations
from zope.annotation.attribute import AttributeAnnotations
from zope.component import getGlobalSiteManager

import colorsys
import mock
import unittest


class MockRequest(object):
    form = {}

    def __init__(self, agent=None):
        self.__headers = {}
        if agent is not None:
            self.__headers['User-Agent'] = agent

    def purge_memoize(self):
        ''' Clean up the memoize cache
        '''
        try:
            self.__annotations__.pop('plone.memoize', None)
        except AttributeError:
            pass

    def get_header(self, key, default):
        return self.__headers.get(key, default)


class MockSession(object):

    def __init__(self, account=None):
        self.account = account


class TestURLs(EuphorieIntegrationTestCase):

    def setUp(self):
        super(TestURLs, self).setUp()
        # Set locals
        request = testRequest()
        locals.request = request
        self.loginAsPortalOwner()
        self.client = self.portal.client
        # Add survey
        self.client._setOb('en', ClientCountry('en'))
        country = self.client['en']
        country._setOb("sector", ClientSector('sector'))
        country["sector"].title = u"Test sector"
        country["sector"].id = "sector"
        sector = country["sector"]
        survey = Survey('survey')
        survey.title = u'Test Survey'
        survey.introduction = u"This is a survey that is well suited for tests"
        survey.language = 'en'
        sector._setOb('survey', survey)

    def testBaseURL(self):
        country = self.client['en']
        survey = self.client['en']['sector']['survey']

        request = testRequest()
        request.client = self.client
        view = WebHelpers(self.client, request)
        self.assertTrue(
            view._base_url().startswith(self.client.absolute_url())
        )
        self.assertFalse(view._base_url().startswith(country.absolute_url()))

        view = WebHelpers(country, testRequest())
        self.assertTrue(view._base_url().startswith(country.absolute_url()))

        request = testRequest()
        view = WebHelpers(survey, testRequest())
        view._survey = survey
        self.assertTrue(view._base_url().startswith(survey.absolute_url()))

        view = WebHelpers(country, testRequest())
        self.assertFalse(view._base_url().startswith(survey.absolute_url()))
        self.assertTrue(view._base_url().startswith(country.absolute_url()))


class WebhelperUnitTests(unittest.TestCase):

    def patch_view(self, name, is_property=False):
        dotted = '.'.join((WebHelpers.__module__, WebHelpers.__name__, name))
        if is_property:
            new_callable = mock.PropertyMock
        else:
            new_callable = None
        return mock.patch(dotted, new_callable=new_callable)

    def test_is_owner(self):
        # If no session is set is_owner return False
        # Allow memoize
        gsm = getGlobalSiteManager()
        gsm.registerAdapter(
            AttributeAnnotations, (MockRequest, ), IAnnotations
        )

        view = WebHelpers(None, MockRequest())
        self.assertEqual(view.session, None)
        self.assertFalse(view.is_owner())
        # Otherwise we will return True is the session account is equal
        # to the current account
        with self.patch_view('session', is_property=True) as mocked_session:
            session = MockSession('account_1')
            mocked_session.return_value = session
            view.get_current_account = lambda: 'account_2'
            self.assertFalse(view.is_owner())
            view.get_current_account = lambda: 'account_1'
            view.request.purge_memoize()
            self.assertTrue(view.is_owner())


class WebhelperTests(EuphorieIntegrationTestCase):

    def _createView(self, agent=None):
        return WebHelpers(self.portal.client, MockRequest(agent))

    def testIsIphone_NoUserAgent(self):
        self.assertEqual(self._createView().is_iphone, False)

    def testIsIphone_IE6(self):
        view = self._createView(
            agent='Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.2; '
            'SV1; .NET CLR 1.1.4322; .NET CLR 2.0.50727; InfoPath.1)'
        )
        self.assertEqual(view.is_iphone, False)

    def testIsIphone_iPhone30(self):
        view = self._createView(
            agent='Mozilla/5.0 (iPod; U; CPU iPhone OS 3_0 like Mac OS X; '
            'en-us) AppleWebKit/528.18 (KHTML, like Gecko) '
            'Version/4.0 Mobile/7A341 Safari/528.16'
        )
        self.assertEqual(view.is_iphone, True)


class HasTextTests(unittest.TestCase):

    def testNone(self):
        self.assertEqual(utils.HasText(None), False)

    def testEmpty(self):
        self.assertEqual(utils.HasText(u''), False)

    def testSimpleText(self):
        self.assertEqual(utils.HasText(u'Hello, World'), True)

    def testWhitespaceOnly(self):
        self.assertEqual(utils.HasText(u'   &nbsp;'), False)

    def testTagsOnly(self):
        self.assertEqual(utils.HasText(u'<strong></strong>'), False)

    def testWhitespaceInTag(self):
        self.assertEqual(utils.HasText(u'<strong>  </strong>'), False)

    def testTagsAndText(self):
        self.assertEqual(utils.HasText(u'<strong>STRONG</strong>'), True)


class RelativePathTests(unittest.TestCase):

    def _createObject(self, id):
        obj = SimpleItem()
        obj.id = id
        return obj

    def testSameItem(self):
        obj = self._createObject('dummy')
        self.assertEqual(utils.RelativePath(obj, obj), '')

    def testDirectChild(self):
        parent = self._createObject('parent')
        child = self._createObject('child').__of__(parent)
        self.assertEqual(utils.RelativePath(parent, child), 'child')

    def testDirectParent(self):
        parent = self._createObject('parent')
        child = self._createObject('child').__of__(parent)
        self.assertEqual(utils.RelativePath(parent, child), 'child')

    def testOtherTree(self):
        parent = self._createObject('root')
        child1 = self._createObject('child1').__of__(parent)
        grandchild = self._createObject('grandchild').__of__(child1)
        child2 = self._createObject('child2').__of__(parent)
        self.assertEqual(
            utils.RelativePath(child2, grandchild), '../child1/grandchild'
        )


class MatchColourTests(unittest.TestCase):

    def from_hls(self, h, l, s):
        (r, g, b) = colorsys.hls_to_rgb(h, l, s)
        return '#%02x%02x%02x' % (r * 255, g * 255, b * 255)

    def to_hls(self, colour):
        (r, g, b) = getrgb(colour)
        return colorsys.rgb_to_hls(r / 255.0, g / 255.0, b / 255.0)

    def testLowLuminosity(self):
        self.assertAlmostEqual(
            self.to_hls(utils.MatchColour(self.from_hls(0, 0, 0)))[1], 0.65, 2
        )

    def testHighLuminosity(self):
        self.assertAlmostEqual(
            self.to_hls(utils.MatchColour(self.from_hls(0, 0.8, 0)))[1], 0.2, 2
        )

    def testYellowishColour(self):
        self.assertAlmostEqual(
            self.to_hls(utils.MatchColour(self.from_hls(0.20, 0, 0)))[1], 0.65,
            2
        )


class RandomStringTests(unittest.TestCase):

    def testOutputChanges(self):
        self.assertNotEquals(utils.randomString(), utils.randomString())

    def testLength(self):
        self.assertEquals(len(utils.randomString(5)), 5)



# coding=utf-8
"""
Solution
--------

A standard Solution that a user can select for a particular risk.
"""

from .. import MessageFactory as _
from euphorie.content import utils
# from euphorie.content.utils import IToolTypesInfo
from five import grok
from plone.directives import dexterity
from plone.directives import form
from plone.indexer import indexer
from plonetheme.nuplone.skin.interfaces import NuPloneSkin
# from z3c.appconfig.interfaces import IAppConfig
from zope import schema
# from zope.component import getUtility
from zope.i18n import translate
from zope.interface import implements


grok.templatedir("templates")


class ISolution(form.Schema):
    """A standard solution for a risk.

    Risk questions can have standard solutions that can be applied in
    most environments.
    """

    description = schema.Text(
        title=_("label_module_description", u"Description"),
        description=_(
            "help_module_description",
            default=u"Include any relevant information that may be "
            u"helpful for the end-user."
        ),
        required=True
    )
    form.order_after(description="title")

    action_plan = schema.Text(
        title=_(
            "label_measure_action_plan",
            default=u"General approach (to eliminate or reduce the risk)"
        ),
        description=_(
            "help_measure_action_plan",
            default=u"Describe your general approach to eliminate or (if "
            u"the risk is not avoidable) reduce the risk."
        ),
        required=True
    )

    prevention_plan = schema.Text(
        title=_(
            "label_measure_prevention_plan",
            default=u"Specific action(s) required to implement this "
            u"approach"
        ),
        description=_(
            "help_measure_prevention_plan",
            default=u"Describe the specific action(s) required to "
            u"implement this approach (to eliminate or to reduce the risk)."
        ),
        required=False
    )

    requirements = schema.Text(
        title=_(
            "label_measure_requirements",
            default=u"Level of expertise and/or requirements needed"
        ),
        description=_(
            "help_measure_requirements",
            default=u'Describe the level of expertise needed to implement '
            u'the measure, for instance "common sense (no OSH knowledge '
            u'required)", "no specific OSH expertise, but minimum OSH '
            u'knowledge or training and/or consultation of OSH guidance '
            u'required", or "OSH expert". You can also describe here '
            u'any other additional requirement (if any).'
        ),
        required=False
    )

    # show_in_identification = schema.Bool(
    #     title=_(
    #         "label_show_in_identification",
    #         default=u"Show this measure during Identification?"),
    #     description=_(
    #         "help_show_in_identification",
    #         default=u"Show this solution during the Identification phase as a "
    #         u"potential measure that is already in place? If the user selects "
    #         u"this measure during Identification, it will no longer be offered"
    #         u" in the Action Plan phase."),
    #     default=True,
    #     required=False,
    # )


class Solution(dexterity.Item):
    implements(ISolution)
    title = _("title_common_solution", default=u"Measure")
    prevention_plan = None
    requirements = None
    portal_type = 'euphorie.solution'

    def Title(self):
        survey = utils.getSurvey(self)
        return translate(
            Solution.title,
            context=self.REQUEST,
            target_language=survey.language
        ).encode('utf-8')


@indexer(ISolution)
def SearchableTextIndexer(obj):
    """ Index the problem_description, question and solution_direction
    """
    return " ".join([
        obj.description, obj.action_plan or '', obj.prevention_plan or '',
        obj.requirements or ''
    ])


class View(grok.View):
    """ View name: @@nuplone-view
    """
    grok.context(ISolution)
    grok.require("zope2.View")
    grok.layer(NuPloneSkin)
    grok.name("nuplone-view")
    grok.template("solution_view")


class Add(dexterity.AddForm):
    grok.context(ISolution)
    grok.name("euphorie.solution")
    grok.require("euphorie.content.AddNewRIEContent")

    def __init__(self, context, request):
        from euphorie.content.survey import get_tool_type
        dexterity.AddForm.__init__(self, context, request)
        # appconfig = getUtility(IAppConfig)
        # settings = appconfig.get('euphorie')
        # self.use_existing_measures = settings.get('use_existing_measures', False)
        # self.tool_type = get_tool_type(context)

    # def updateWidgets(self):
    #     super(Add, self).updateWidgets()
    #     tt = getUtility(IToolTypesInfo)
    #     if not (
    #         self.use_existing_measures and
    #         self.tool_type in tt.types_existing_measures
    #     ):
    #         self.widgets["show_in_identification"].mode = "hidden"


class Edit(form.SchemaEditForm):
    grok.context(ISolution)
    grok.require("cmf.ModifyPortalContent")
    grok.layer(NuPloneSkin)
    grok.name("edit")

    # def __init__(self, context, request):
    #     from euphorie.content.survey import get_tool_type
    #     appconfig = getUtility(IAppConfig)
    #     settings = appconfig.get('euphorie')
    #     self.use_existing_measures = settings.get('use_existing_measures', False)
    #     self.tool_type = get_tool_type(context)
    #     form.SchemaEditForm.__init__(self, context, request)

    # def updateWidgets(self):
    #     super(Edit, self).updateWidgets()
    #     tt = getUtility(IToolTypesInfo)
    #     if not (
    #         self.use_existing_measures and
    #         self.tool_type in tt.types_existing_measures
    #     ):
    #         self.widgets["show_in_identification"].mode = "hidden"



"""
Session
-------

Create and update sessions.
"""

from Acquisition import aq_inner
from Acquisition import aq_parent
from euphorie.client import model
from euphorie.client.cookie import deleteCookie
from euphorie.client.cookie import getCookie
from euphorie.client.cookie import setCookie
from euphorie.client.utils import getRequest
from euphorie.client.utils import getSecret
from plone import api
from z3c.saconfig import Session

import logging


log = logging.getLogger(__name__)
SESSION_COOKIE = "_eu_session"


class SessionManagerFactory(object):
    """Session management handling for the client.

    Never use this class directly: instead use the global
    :py:data:`SessionManager` instance.
    """
    model = model.SurveySession

    @property
    def session(self):
        """The current active client session. If no session is active None
        is returned.

        :rtype: :py:class:`self.model` or None
        """
        request = getRequest()
        if "euphorie.session" in request.other:
            return request.other["euphorie.session"]

        id = self.id
        if id is None:
            return None

        session = Session.query(self.model).get(id)
        request.other["euphorie.session"] = session
        return session

    def get_session_by_id(self, sessionid=None):
        if sessionid is None:
            return self.session
        return Session.query(self.model).get(sessionid)

    def start(self, title, survey, account=None):
        """Create a new session and activate it.

        :param title: title for the new session.
        :type title: unicode
        :param survey: survey for which the session is being created
        :type survey: :py:class:`euphorie.content.survey.Survey`
        :rtype: :py:class:`euphorie.client.model.SurveySession` instance
        """
        survey_session = create_survey_session(title, survey, account)
        request = getRequest()
        setCookie(
            request.response,
            getSecret(),
            SESSION_COOKIE,
            survey_session.id,
        )
        request.other['euphorie.session'] = survey_session
        return survey_session

    @property
    def webhelpers(self):
        return api.content.get_view(
            'webhelpers',
            context=api.portal.get(),
            request=getRequest(),
        )

    def resume(self, session):
        """Activate the given session.

        :param session: session to activate
        :type session: :py:class:`euphorie.client.model.SurveySession`
        """
        return self.webhelpers.resume(session)

    def stop(self):
        """End the current active session.
        """
        request = getRequest()
        if "euphorie.session" in request.other:
            del request.other["euphorie.session"]
        deleteCookie(request.response, SESSION_COOKIE)
        setCookie(request.response, getSecret(), SESSION_COOKIE, '')

    @property
    def id(self):
        """The id of the current session, or None if there is no active
        session.

        This method does not perform any security checks.

        :rtype: int or None
        """
        request = getRequest()
        if "euphorie.session" in request.other:
            return request.other["euphorie.session"].id

        session_id = getCookie(request, getSecret(), "_eu_session")

        try:
            return int(session_id)
        except TypeError:
            return None


SessionManager = SessionManagerFactory()
"""Global instance of :py:class:`SessionManagerFactory`.
"""


def create_survey_session(title, survey, account=None, **params):
    """Create a new survey session.

    :param title: title for the new session.
    :type title: unicode
    :param survey: survey for which the session is being created
    :type survey: :py:class:`euphorie.content.survey.Survey`
    :rtype: :py:class:`euphorie.client.model.SurveySession` instance
    """
    if account is None:
        account = model.get_current_account()

    session = Session()
    sector = aq_parent(aq_inner(survey))
    country = aq_parent(sector)
    zodb_path = '%s/%s/%s' % (country.id, sector.id, survey.id)
    survey_session = SessionManager.model(
        title=title,
        zodb_path=zodb_path,
        account_id=account.id,
        group_id=account.group_id,
    )
    for key in params:
        setattr(survey_session, key, params[key])
    session.add(survey_session)
    session.refresh(account)
    session.flush()  # flush so we get a session id
    return survey_session


__all__ = ["SessionManager"]



# coding=utf-8
from Acquisition import aq_parent
from euphorie.content.tests.utils import createSector
from euphorie.testing import EuphorieIntegrationTestCase


class CountryTests(EuphorieIntegrationTestCase):
    def createCountry(self):
        sector = createSector(self.portal, login="sector")
        return aq_parent(sector)

    def testCanNotBeCopied(self):
        country = self.createCountry()
        self.assertFalse(country.cb_isCopyable())



# coding=utf-8
from AccessControl import getSecurityManager
from Acquisition import aq_inner
from anytree import NodeMixin
from anytree.node.util import _repr
from collections import defaultdict
from collections import OrderedDict
from datetime import datetime
from euphorie import MessageFactory as _
from euphorie.client import model
from euphorie.client import utils
from euphorie.client.country import IClientCountry
from euphorie.client.model import get_current_account
from euphorie.client.model import Group
from euphorie.client.model import Module
from euphorie.client.model import Risk
from euphorie.client.model import SurveySession
from euphorie.client.model import SurveyTreeItem
from euphorie.client.sector import IClientSector
from euphorie.client.session import SessionManager
from euphorie.content.survey import ISurvey
from logging import getLogger
from plone import api
from plone.memoize.view import memoize
from plone.memoize.view import memoize_contextless
from Products.CMFPlone.utils import safe_unicode
from Products.Five import BrowserView
from z3c.saconfig import Session
from zExceptions import Unauthorized
from zope.event import notify
from zope.i18n import translate
from zope.lifecycleevent import ObjectModifiedEvent

import six


logger = getLogger(__name__)


class Node(NodeMixin):
    def __init__(self, context, parent=None, **kwargs):
        self.__dict__.update(kwargs)
        self.context = context
        self.parent = parent

    @property
    def groups(self):
        """ Assume childrens are groups and return them sorted by title
        """
        return sorted(self.children, key=lambda x: x.title)

    @property
    def sessions(self):
        """ Assume childrens are sessions and return them sorted by
        reversed modification date
        """
        return sorted(self.children, key=lambda x: x.context.modified, reverse=True)

    @property
    def survey_templates(self):
        """ Return all children that are survey_templates, sorted by title
        """
        return sorted(
            [item for item in self.children if item.type == "survey_template"],
            key=lambda x: x.title.lower(),
        )

    @property
    def categories(self):
        """ Return all children that are categories, sorted by title
        """
        return sorted(
            [item for item in self.children if item.type == "category"],
            key=lambda x: x.title,
        )

    def __repr__(self):
        args = [
            "%r"
            % self.separator.join([""] + [repr(node.context) for node in self.path])
        ]
        return _repr(self, args=args, nameblacklist=["context"])


class SessionsView(BrowserView):

    # switch from radio buttons to dropdown above this number of tools
    tools_threshold = 12
    variation_class = "variation-dashboard"

    @property
    @memoize
    def my_context(self):
        if IClientCountry.providedBy(self.context):
            return "country"
        elif ISurvey.providedBy(self.context):
            return "survey"

    @property
    @memoize_contextless
    def portal(self):
        """ The currenttly authenticated account
        """
        return api.portal.get()

    @property
    @memoize_contextless
    def account(self):
        """ The currenttly authenticated account
        """
        return model.get_current_account()

    @property
    @memoize_contextless
    def scope_options(self):
        """ We have the possibility to display only the sessions
        that were made by this user or all the accessible sessions through
        the group ownership
        """
        account = self.account
        if not account or not account.group:
            return []
        options = [
            {"value": "mine", "label": _("Show my risk assessments only")},
            {"value": "all", "label": _("Show all risk assessments")},
        ]
        selected = self.request.get("scope")
        for option in options:
            if option["value"] == selected:
                option["selected"] = "selected"
        return options

    @memoize
    def get_survey_by_path(self, zodb_path):
        return self.context.restrictedTraverse(six.binary_type(zodb_path), None)

    @property
    @memoize
    def sessions_root(self):
        return Node(None, title="", type="root")

    @memoize
    def get_group_node(self, group):
        """ Get the Node for this group.
        """
        if group is None:
            # Everything is grouped under the sessions_root node
            return self.sessions_root
        return Node(
            group,
            parent=self.get_group_node(group.parent),
            title=group.fullname,
            type="department",
        )

    @memoize
    def get_survey_node(self, survey, group):
        """ Get a node for this survey, it might be in a group
        """
        return Node(
            survey, parent=self.get_group_node(group), title=survey.title, type="tool"
        )

    @memoize
    def get_session_node(self, session):
        """ Get a node for this session
        """
        group = session.group
        survey = self.get_survey_by_path(session.zodb_path)
        if not survey:
            return
        return Node(
            session,
            parent=self.get_survey_node(survey, group),
            title=session.title,
            type="session",
        )

    def filter_valid_sessions(self, sessions):
        return [
            session
            for session in sessions
            if self.get_survey_by_path(session.zodb_path)
        ]

    @memoize
    def get_sessions(self):
        """ Given some sessions create a tree
        """
        scope = self.request.get("scope")
        account = self.account
        if not account:
            sessions = []
        elif scope == "all":
            sessions = self.account.sessions + self.account.acquired_sessions
        else:
            sessions = self.account.sessions

        return self.filter_valid_sessions(sessions)

    @memoize
    def get_ordered_sessions(self):
        """ Given some sessions create a tree
        """
        return sorted(self.get_sessions(), key=lambda x: x.modified, reverse=True)

    @memoize
    def get_sessions_tree_root(self):
        """ Given some sessions create a tree
        """
        sessions = self.get_sessions()
        map(self.get_session_node, sessions)
        return self.sessions_root

    # Here, we assemble the list of available tools for starting a new session

    @property
    @memoize
    def survey_templates_root(self):
        return Node(None, title="", type="root")

    @memoize
    def get_survey_category_node(self, category):
        if category is None:
            # Everything is grouped under the survey_templates_root node
            return self.survey_templates_root
        return Node(
            category,
            parent=self.get_survey_category_node(None),
            title=category,
            type="category",
        )

    @memoize
    def get_survey_template_node(self, survey_item):
        category, survey, id = survey_item
        return Node(
            survey,
            parent=self.get_survey_category_node(category),
            title=survey.title,
            type="survey_template",
            id=id,
        )

    @memoize
    def get_survey_templates_tree_root(self):
        survey_items = self.get_survey_templates()
        map(self.get_survey_template_node, survey_items)
        return self.survey_templates_root

    @memoize
    def get_survey_templates(self):
        # This number might be higher than the actual amount of survey
        # templates, since some might appear under more than one category.
        self.template_count = 0
        # this is a list of tuples of the form
        # (category name, survey object, survey id)
        survey_items = []
        language = self.request.locale.id.language or ""
        for sector in aq_inner(self.context).values():
            if not IClientSector.providedBy(sector):
                continue

            for survey in sector.values():
                if not ISurvey.providedBy(survey):
                    continue
                if getattr(survey, "preview", False):
                    continue
                if (
                    survey.language
                    and survey.language != language
                    and not survey.language.strip().startswith(language)
                ):
                    continue
                if getattr(survey, "obsolete", False):
                    continue
                categories = getattr(survey, "tool_category", None)
                id = "%s/%s" % (sector.id, survey.id)
                if not isinstance(categories, list):
                    categories = [categories]
                if not categories:
                    categories = [None]
                for category in categories:
                    survey_items.append((category, survey, id))
                    self.template_count += 1
        return sorted(survey_items, key=lambda x: (x[0], x[1].title))

    def _updateSurveys(self):
        self.surveys = []
        self.obsolete_surveys = []

        language = self.request.locale.id.language or ""
        for sector in aq_inner(self.context).values():
            if not IClientSector.providedBy(sector):
                continue

            for survey in sector.values():
                if not ISurvey.providedBy(survey):
                    continue
                if getattr(survey, "preview", False):
                    continue
                if (
                    survey.language
                    and survey.language != language
                    and not survey.language.strip().startswith(language)
                ):
                    continue
                info = {"id": "%s/%s" % (sector.id, survey.id), "title": survey.title}
                if getattr(survey, "obsolete", False):
                    # getattr needed for surveys which were published before
                    # the obsolete flag added.
                    self.obsolete_surveys.append(info)
                else:
                    self.surveys.append(info)
        self.surveys.sort(key=lambda s: s["title"])
        self.obsolete_surveys.sort(key=lambda s: s["title"])

    def _NewSurvey(self, info, account=None):
        """Utility method to start a new survey session."""
        context = aq_inner(self.context)
        survey = info.get("survey")
        survey = context.restrictedTraverse(survey)
        if not ISurvey.providedBy(survey):
            logger.error("Tried to start invalid survey %r", info.get("survey"))
            # Things are sufficiently messed up at this point that rendering
            # breaks, so trigger a redirect to the same URL again.
            self.request.response.redirect(context.absolute_url())
            return
        title = info.get("title", u"").strip()
        if not title:
            title = survey.Title()

        SessionManager.start(title=title, survey=survey, account=account)
        self.request.response.redirect(
            "%s/start?initial_view=1&new_session=1" % survey.absolute_url()
        )

    def _ContinueSurvey(self, info):
        """Utility method to continue an existing session."""
        session = Session.query(model.SurveySession).get(info["session"])
        SessionManager.resume(session)
        survey = self.request.client.restrictedTraverse(
            six.binary_type(session.zodb_path)
        )
        extra = ""
        if info.get("new_clone", None):
            extra = "&new_clone=1"
        self.request.response.redirect(
            "%s/resume?initial_view=1%s" % (survey.absolute_url(), extra)
        )

    def tool_byline(self):
        lang = getattr(self.request, "LANGUAGE", "en")
        title = translate(
            _("title_tool", default=u"OiRA - Online interactive Risk Assessment"),
            target_language=lang,
        )
        return title.split("-")[-1].strip()

    def __call__(self):
        if not self.account:
            raise Unauthorized()

        utils.setLanguage(self.request, self.context)
        reply = self.request.form
        action = reply.get("action")
        if action == "new":
            if self.my_context == "survey":
                reply["survey"] = "/".join(self.context.getPhysicalPath())
            return self._NewSurvey(reply)
        elif action == "continue":
            return self._ContinueSurvey(reply)
        self._updateSurveys()
        return super(SessionsView, self).__call__()


class SessionBrowserNavigator(SessionsView):
    """ Logic to build the navigator for the sessions
    """

    group_model = Group
    no_splash = True

    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    @property
    @memoize
    def groupid(self):
        return self.request.get("groupid")

    @memoize
    def get_root_group(self, groupid=None):
        """ Return the group that is the root of the navigation tree
        """
        if not groupid:
            groupid = self.groupid
        if not groupid:
            return
        base_query = Session.query(self.group_model).order_by(
            self.group_model.short_name
        )
        return base_query.filter(self.group_model.group_id == groupid).one()

    @property
    @memoize
    def searchable_text(self):
        """ Return the text we need to search in postgres
        already surrounded with '%'
        Allow a minimum size of 3 characters to reduce the load.
        """
        searchable_text = self.request.get("SearchableText")
        if not isinstance(searchable_text, six.string_types):
            return ""
        if len(searchable_text) < 3:
            return ""
        return u"%{}%".format(safe_unicode(searchable_text))

    def get_tools_tree(self, zodb_path=None):
        """ Return a dict like structure to render the leaf sessions,
        something like:

        {
            'tool': [<SurveySession>, ...],
            ...
        }
        Optionally, we can pass in zodb_path, to filter for tools that match
        this path
        """
        sessions = self.leaf_sessions()
        if not sessions.count():
            return {}
        tools = defaultdict(list)
        for session in sessions:
            # XXX
            # Filter by zodb_path for specific tools
            # if zodb_path and session.zodb_path != zodb_path:
            #     continue
            tool = self.get_survey_by_path(session.zodb_path)
            tools[tool].append(session)

        ordered_tools = OrderedDict()
        for tool in sorted(
            [x for x in tools.keys() if x], key=lambda s: s.title.lower()
        ):
            ordered_tools[tool] = sorted(
                tools[tool], key=lambda s: s.modified, reverse=True
            )

        return ordered_tools

    @memoize
    def leaf_groups(self, groupid=None):
        """ Nothing to do in main OiRA - to be filled in customer-specific
        packages.
        Here we just return a Query with 0 items.
        """
        base_query = Session.query(SurveySession)
        return base_query.filter(False)

    def leaf_sessions(self):
        """ The sessions we want to display in the navigation
        """
        base_query = Session.query(SurveySession).order_by(SurveySession.title)
        if self.searchable_text:
            base_query = base_query.filter(
                SurveySession.title.ilike(self.searchable_text)
            )
        account = get_current_account()
        return base_query.filter(SurveySession.account_id == account.id)

    def has_content(self):
        """ Checks if we have something meaningfull to display
        """
        if self.leaf_groups().count():
            return True
        if len(self.get_tools_tree()):
            return True
        return False


class ConfirmationDeleteSession(BrowserView):
    """View name: @@confirmation-delete-session
    """

    @property
    @memoize_contextless
    def webhelpers(self):
        return api.content.get_view("webhelpers", api.portal.get(), self.request)

    @property
    @memoize_contextless
    def session_title(self):
        try:
            self.session_id = int(self.request.get("id"))
        except (ValueError, TypeError):
            raise KeyError("Invalid session id")
        session = (
            Session.query(SurveySession)
            .filter(SurveySession.id == self.session_id)
            .one()
        )
        if session is None:
            raise KeyError("Unknown session id")
        if not self.webhelpers.can_delete_session(session):
            raise Unauthorized()
        return session.title

    def __call__(self, *args, **kwargs):
        """ Before rendering check if we can find session title
        """
        self.session_title
        self.no_splash = True
        return super(ConfirmationDeleteSession, self).__call__(*args, **kwargs)


def sql_clone(obj, skip={}, session=None):
    """ Clone a sql object avoiding the properties in the skip parameter

    The skip parameter is optional but you probably want to always pass the
    primary key
    """
    # Never copy the _sa_instance_state attribute
    skip.add("_sa_instance_state")
    params = {key: value for key, value in obj.__dict__.iteritems() if key not in skip}
    clone = obj.__class__(**params)
    if session:
        session.add(clone)
    return clone


class CloneSession(BrowserView):
    """View name: @@confirmation-clone-session
    """

    @property
    @memoize
    def webhelpers(self):
        return api.content.get_view("webhelpers", self.context, self.request)

    @property
    @memoize
    def session(self):
        """ Get the requested session
        """
        try:
            session_id = int(self.request.get("id"))
        except (ValueError, TypeError):
            raise KeyError("Invalid session id")
        return Session.query(SurveySession).filter(SurveySession.id == session_id).one()

    def get_cloned_session(self):
        sql_session = Session
        old_session = self.session
        new_session = sql_clone(
            old_session,
            skip={
                "id",
                "created",
                "modified",
                "last_modifier_id",
                "company",
                "published",
                "group_id",
            },
            session=sql_session,
        )
        lang = getattr(self.request, "LANGUAGE", "en")
        new_session.title = u"{}: {}".format(
            translate(_("prefix_cloned_title", default=u"COPY"), target_language=lang),
            new_session.title,
        )
        account = self.webhelpers.get_current_account()
        new_session.group = account.group
        new_session.modified = new_session.created = datetime.now()
        new_session.account = account
        if old_session.company:
            new_session.company = sql_clone(
                old_session.company, skip={"id", "session"}, session=sql_session
            )

        risk_module_skipped_attributes = {
            "id",
            "session",
            "sql_module_id",
            "parent_id",
            "session_id",
            "sql_risk_id",
            "risk_id",
        }
        module_mapping = {}

        old_modules = sql_session.query(Module).filter(
            SurveyTreeItem.session == old_session
        )
        for old_module in old_modules:
            new_module = sql_clone(
                old_module, skip=risk_module_skipped_attributes, session=sql_session
            )
            module_mapping[old_module.id] = new_module
            new_module.session = new_session

        old_risks = sql_session.query(Risk).filter(
            SurveyTreeItem.session == old_session
        )
        for old_risk in old_risks:
            new_risk = sql_clone(
                old_risk, skip=risk_module_skipped_attributes, session=sql_session
            )
            new_risk.parent_id = module_mapping[old_risk.parent_id].id
            new_risk.session = new_session

            for old_plan in old_risk.action_plans:
                new_plan = sql_clone(
                    old_plan, skip={"id", "risk_id"}, session=sql_session
                )
                new_plan.risk = new_risk
        notify(ObjectModifiedEvent(new_session))
        return new_session

    def clone(self):
        """ Clone this session and redirect to the start view
        """
        new_session = self.get_cloned_session()
        api.portal.show_message(
            _("The risk assessment has been cloned"), self.request, "success"
        )
        target = "{contexturl}/@@view?action=continue&new_clone=1&session={sessionid}".format(  # noqa: E501
            contexturl=self.context.absolute_url(), sessionid=new_session.id
        )
        self.request.response.redirect(target)



# coding=utf-8
from zope.interface import implementer
from plone.transformchain.interfaces import ITransform
from plone.app.theming.transform import ThemeTransform
# from lxml import etree


@implementer(ITransform)
class EuphorieTransform(ThemeTransform):
    """ A transform that comes after plone.app.theming.transform and before
    plone.protect.auto
    Since theming is not used in Euphorie, the only purpose of this transform
    is to ensure that we wrap the result into an XMLSerializer, so that the
    plone.protext transform does not mangle it. See:
    https://community.plone.org/t/where-does-url-quoting-of-links-happen/6643
    """

    order = 8870

    def transformIterable(self, result, encoding):
        """ Just make sure to call parseTree in order to wrap the result
        """
        result = self.parseTree(result)
        if result is None:
            return None
        # This is the attempt to make it work in the client, because
        # there we only have doctype html (HTML5).
        # However, this causes all pages in the client to be blank,
        # even though in source view it looks fine.
        # More debugging needed.
        # Until then, force doctype XMHTL on the risk_actionplan template,
        # the only location where this problem affects us negatively.
        # if (
        #     result.tree.docinfo.doctype and (
        #         'XHTML' in result.tree.docinfo.doctype or
        #         'html' in result.tree.docinfo.doctype
        #     )
        # ):
        #     result.serializer = etree.tostring
        return result



# coding=utf-8
from euphorie.content.tests.utils import addSurvey
from euphorie.content.tests.utils import BASIC_SURVEY
from euphorie.content.tests.utils import createSector
from euphorie.testing import EuphorieFunctionalTestCase

import mock


class LibraryTests(EuphorieFunctionalTestCase):
    library = [{
        'title': u'Library Sector Title',
        'url': 'http://localhost/sector',
        'surveys': [{
            'title': u'Library Survey Title',
            'path': '/portal/sector/survey/version',
            'url': 'http://localhost/portal/sector/survey/version',
            'portal_type': 'euphorie.sector',
            'children': [{
                'title': u'Library Risk',
                'number': '1',
                'path': '/portal/sector/surveys/version/1',
                'url': 'http://localhost/portal/sector/surveys/version/1',
                'disabled': False,
                'portal_type': 'euphorie.risk',
                'children': []
            }]
        }]
    }]

    def test_render(self):
        self.loginAsPortalOwner()
        sector = createSector(self.portal)
        survey = addSurvey(sector, BASIC_SURVEY)
        browser = self.get_browser(logged_in=True)
        browser.handleErrors = False
        with mock.patch(
            'euphorie.content.library.get_library', return_value=self.library
        ):
            browser.open(survey.absolute_url() + '/@@library')
        assert 'Library Sector Title' in browser.contents
        assert 'Library Survey Title' in browser.contents
        assert 'Library Risk' in browser.contents



# coding=utf-8
"""
Risk
----

Views for the identification and action plan phases.
"""

from Acquisition import aq_chain
from Acquisition import aq_inner
from Acquisition import aq_parent
from collections import OrderedDict
from euphorie import MessageFactory as _
from euphorie.client import model
from euphorie.client.interfaces import IItalyActionPlanPhaseSkinLayer
from euphorie.client.interfaces import IItalyIdentificationPhaseSkinLayer
from euphorie.client.navigation import FindNextQuestion
from euphorie.client.navigation import FindPreviousQuestion
from euphorie.client.navigation import getTreeData
from euphorie.client.navigation import QuestionURL
from euphorie.client.session import SessionManager
from euphorie.client.update import redirectOnSurveyUpdate
from euphorie.client.utils import HasText
from euphorie.content.risk import IRisk
from euphorie.content.solution import ISolution
from euphorie.content.survey import get_tool_type
from euphorie.content.survey import ISurvey
from euphorie.content.utils import IToolTypesInfo
from htmllaundry import StripMarkup
from json import dumps
from json import loads
from plone.memoize.instance import memoize
from Products.CMFPlone.utils import safe_unicode
from Products.Five import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.statusmessages.interfaces import IStatusMessage
from sqlalchemy import and_
from z3c.appconfig.interfaces import IAppConfig
from z3c.appconfig.utils import asBool
from z3c.saconfig import Session
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.i18n import translate
import datetime

IMAGE_CLASS = {
    0: '',
    1: 'twelve',
    2: 'six',
    3: 'four',
    4: 'three',
}


class IdentificationView(BrowserView):
    """A view for displaying a question in the identification phase
    """
    template = ViewPageTemplateFile('templates/risk_identification.pt')
    variation_class = "variation-risk-assessment"

    question_filter = None

    # default value is False, can be overwritten by certain conditions
    skip_evaluation = False

    monitored_properties = {
        "identification": None,
        "postponed": None,
        "frequency": None,
        "severity": None,
        "priority": None,
        "comment": u"",
        "existing_measures": u"[]",
        "training_notes": None,
        "custom_description": None,
    }

    @property
    @memoize
    def webhelpers(self):
        return self.context.restrictedTraverse('webhelpers')

    def __call__(self):
        # Render the page only if the user has edit rights,
        # otherwise redirect to the start page of the session.
        if not (
            self.webhelpers.can_edit_session()
        ):
            return self.request.response.redirect(
                self.webhelpers.survey_url() + '/@@start'
            )
        if redirectOnSurveyUpdate(self.request):
            return

        if self.is_custom_risk:
            # # Fetch the custom_risks module, so that we can at least
            # # traverse to the survey
            # path = "/".join(self.context.zodb_path.split('/')[:-1])
            self.risk = None
        else:
            self.risk = self.request.survey.restrictedTraverse(
                self.context.zodb_path.split("/"))

        appconfig = getUtility(IAppConfig)
        settings = appconfig.get('euphorie')
        self.tti = getUtility(IToolTypesInfo)
        self.my_tool_type = get_tool_type(self.context)
        self.use_existing_measures = (
            asBool(settings.get('use_existing_measures', False)) and
            self.my_tool_type in self.tti.types_existing_measures
        )
        self.use_training_module = asBool(
            settings.get('use_training_module', False))

        if self.request.environ["REQUEST_METHOD"] == "POST":
            reply = self.request.form
            # Don't persist anything if the user skipped the question
            if reply.get("next", None) == 'skip':
                return self.proceed_to_next(reply)
            old_values = {}
            for prop, default in self.monitored_properties.items():
                if prop == "existing_measures":
                    val = dumps([
                        entry for entry in loads(
                            getattr(self.context, prop, default)
                        ) if entry[1]
                    ])
                else:
                    val = getattr(self.context, prop, default)
                old_values[prop] = val
            answer = reply.get("answer", None)
            # If answer is not present in the request, do not attempt to set
            # any answer-related data, since the request might have come
            # from a sub-form.
            if answer:
                self.context.comment = reply.get("comment")
                self.context.postponed = (answer == "postponed")
                if self.context.postponed:
                    self.context.identification = None
                else:
                    self.context.identification = answer
                    if getattr(self.risk, "type", "") in ('top5', 'policy'):
                        self.context.priority = 'high'
                    elif getattr(
                        self.risk, "evaluation_method", ""
                    ) == 'calculated':
                        self.calculatePriority(self.risk, reply)
                    elif (
                        self.risk is None or
                        self.risk.evaluation_method == "direct"
                    ):
                        self.context.priority = reply.get("priority")

            if (
                self.use_existing_measures and
                reply.get('handle_measures_in_place')
            ):
                measures = self.get_existing_measures()
                new_measures = []
                seen = []
                for i, entry in enumerate(measures):
                    on = int(bool(reply.get('measure-{}'.format(i))))
                    new_measures.append((entry[0], on))
                    if on:
                        seen.append(i)
                for k, val in reply.items():
                    if (
                        k.startswith('new-measure') and
                        isinstance(val, str) and
                        val.strip() != ''
                    ):
                        new_measures.append((val, 1))
                    elif k.startswith('present-measure') and val.strip() != '':
                        idx = k.rsplit("-", 1)[-1]
                        try:
                            idx = int(idx)
                        except (TypeError, ValueError):
                            continue
                        if idx in seen:
                            new_measures[idx] = (val, 1)

                # Only save the measures that are active
                self.context.existing_measures = safe_unicode(
                    dumps([entry for entry in new_measures if entry[1]]))

            if self.use_training_module and reply.get('handle_training_notes'):
                self.context.training_notes = reply.get("training_notes")

            # This only happens on custom risks
            if reply.get("handle_custom_description"):
                self.context.custom_description = reply.get("custom_description")

            if reply.get("title"):
                self.context.title = reply.get("title")

            # Check if there was a change. If yes, touch the session
            changed = False
            for prop, default in self.monitored_properties.items():
                if prop == "existing_measures":
                    val = dumps([
                        entry for entry in loads(
                            getattr(self.context, prop, default)
                        ) if entry[1]
                    ])
                else:
                    val = getattr(self.context, prop, None)
                if (
                    val and val != old_values[prop]
                ):
                    changed = True
                    break
            if changed:
                SessionManager.session.touch()

            return self.proceed_to_next(reply)

        else:
            self._prepare_risk()
            if self.is_custom_risk:
                template = ViewPageTemplateFile(
                    'templates/risk_identification_custom.pt').__get__(self, "")  # noqa
                next = FindNextQuestion(
                    self.context,
                    filter=self.question_filter)
                self.has_next_risk = next or False
            else:
                template = self.template
            return template()

    def _prepare_risk(self):
        self.tree = getTreeData(self.request, self.context)
        self.title = self.context.parent.title
        has_risk_description = (
            (self.risk and HasText(self.risk.description)) or
            getattr(self.context, 'custom_description', '')
        )
        self.show_info = (
            getattr(self.risk, "image", None) or
            (
                self.risk is None or
                HasText(self.risk.description)
            )
        )

        number_images = getattr(self.risk, 'image', None) and 1 or 0
        if number_images:
            for i in range(2, 5):
                number_images += getattr(
                    self.risk, 'image{0}'.format(i), None) and 1 or 0
        self.has_images = number_images > 0
        self.number_images = number_images
        self.image_class = IMAGE_CLASS[number_images]
        number_files = 0
        for i in range(1, 5):
            number_files += getattr(
                self.risk, 'file{0}'.format(i), None) and 1 or 0
        self.has_files = number_files > 0
        self.has_legal = HasText(getattr(self.risk, "legal_reference", None))
        self.show_resources = self.has_legal or self.has_files

        self.risk_number = self.context.number

        self.description_probability = _(
            u"help_default_probability", default=u"Indicate how "
            u"likely occurence of this risk is in a normal situation.")
        self.description_frequency = _(
            u"help_default_frequency", default=u"Indicate how often this "
            u"risk occurs in a normal situation.")
        self.description_severity = _(
            u"help_default_severity", default=u"Indicate the "
            u"severity if this risk occurs.")

        tool_types = self.tti()
        tt_default = self.tti.default_tool_type
        tool_type_data = tool_types.get(
            self.my_tool_type, tool_types[tt_default])
        default_type_data = tool_types['classic']
        self.show_existing_measures = False

        # Fill some labels with default texts
        self.answer_yes = default_type_data['answer_yes']
        self.answer_no = default_type_data['answer_no']
        self.answer_na = default_type_data['answer_na']
        self.intro_extra = ""
        if self.is_custom_risk:
            self.intro_extra = tool_type_data.get('custom_intro_extra', '')
            if self.use_existing_measures:
                self.answer_yes = tool_type_data['answer_yes']
                self.answer_no = tool_type_data['answer_no']
        self.button_add_extra = tool_type_data.get('button_add_extra', '')
        self.intro_questions = tool_type_data.get('intro_questions', '')
        self.placeholder_add_extra = tool_type_data.get(
                'placeholder_add_extra', '')
        self.button_remove_extra = ""
        if self.use_existing_measures:
            measures = self.get_existing_measures()
            # Only show the form to select and add existing measures if
            # at least one pre-existring measure is present
            # In this case, also change some labels
            if len(measures):
                self.show_existing_measures = True
                self.intro_extra = tool_type_data.get('intro_extra', '')
                self.button_remove_extra = tool_type_data.get(
                    'button_remove_extra', '')
                self.answer_yes = tool_type_data['answer_yes']
                self.answer_no = tool_type_data['answer_no']
                self.answer_na = tool_type_data['answer_na']
            if not self.context.existing_measures:
                existing_measures = [
                    (text, 0) for text in measures
                ]
                self.context.existing_measures = safe_unicode(
                    dumps(existing_measures))

        if getattr(self.request.survey, 'enable_custom_evaluation_descriptions', False):
            if self.request.survey.evaluation_algorithm != 'french':
                custom_dp = getattr(
                    self.request.survey, 'description_probability', '') or ''
                self.description_probability = custom_dp.strip() or self.description_probability
            custom_df = getattr(self.request.survey, 'description_frequency', '') or ''
            self.description_frequency = custom_df.strip() or self.description_frequency
            custom_ds = getattr(self.request.survey, 'description_severity', '') or ''
            self.description_severity = custom_ds.strip() or self.description_severity

        # compute training side template
        self.slide_template = (
            (has_risk_description or number_images) and "template-two-column"
            or "template-default"
        )

        # Italian special
        if IItalyIdentificationPhaseSkinLayer.providedBy(self.request):
            self.skip_evaluation = True

    def proceed_to_next(self, reply):
        if reply.get("next", None) == "previous":
            next = FindPreviousQuestion(
                self.context,
                filter=self.question_filter)
            if next is None:
                # We ran out of questions, step back to intro page
                url = "%s/identification" % \
                    self.request.survey.absolute_url()
                self.request.response.redirect(url)
                return
        elif reply.get("next", None) in ("next", "skip"):
            next = FindNextQuestion(
                self.context,
                filter=self.question_filter)
            if next is None:
                # We ran out of questions, proceed to the action plan
                url = "%s/actionplan" % self.request.survey.absolute_url()
                self.request.response.redirect(url)
                return

        elif reply.get("next", None) == "add_custom_risk":
            module = aq_parent(self.context)
            sql_module_q = Session.query(model.Module).filter(
                and_(
                    model.SurveyTreeItem.session == SessionManager.session,
                    model.Module.zodb_path == u'custom-risks',
                    )
            )
            if not sql_module_q.count():
                url = QuestionURL(
                    self.request.survey, self.context, phase="identification")
                self.request.response.redirect(url)
                return
            sql_module = sql_module_q.one()
            view = getMultiAdapter(
                (sql_module, self.request), name="index_html")
            risk_id = view.add_custom_risk()
            url = "%s/%d" % (module.absolute_url(), risk_id)
            self.request.response.redirect(url)
            return
        elif reply.get("next", None) == "actionplan":
            url = "%s/actionplan" % self.request.survey.absolute_url()
            self.request.response.redirect(url)
            return
        # stay on current risk
        else:
            next = self.context

        url = QuestionURL(self.request.survey, next, phase="identification")
        self.request.response.redirect(url)

    def get_existing_measures(self):
        if not self.risk:
            defined_measures = []
        else:
            defined_measures = (
                self.risk.get_pre_defined_measures(self.request) or "")
        try:
            saved_existing_measures = loads(
                self.context.existing_measures or "")
            # Backwards compat. We used to save dicts before we
            # switched to list of tuples.
            if isinstance(saved_existing_measures, dict):
                saved_existing_measures = [
                    (k, v) for (k, v) in saved_existing_measures.items()]

            saved_measure_texts = OrderedDict()
            for text, on in saved_existing_measures:
                saved_measure_texts.update({text: on})

            existing_measures = []
            # All the pre-defined measures are always shown, either
            # activated or deactivated
            for text in defined_measures:
                active = saved_measure_texts.get(text)
                if active is not None:
                    saved_measure_texts.pop(text)
                    existing_measures.append((text, active))
                else:
                    existing_measures.append((text, 0))

            # Finally, add the user-defined measures as well
            for text, on in saved_measure_texts.items():
                existing_measures.append((text, on))
        except ValueError:
            existing_measures = [
                (text, 0) for text in defined_measures
            ]
            self.context.existing_measures = safe_unicode(
                dumps(existing_measures))
        return existing_measures

    @property
    def use_problem_description(self):
        risk = self.request.survey.restrictedTraverse(
            self.context.zodb_path.split("/"))
        text = risk.problem_description
        return bool(text and text.strip())

    @property
    def is_custom_risk(self):
        return getattr(self.context, 'is_custom_risk', False)

    def evaluation_algorithm(self, risk):
        return evaluation_algorithm(risk)

    def calculatePriority(self, risk, reply):
        self.context.frequency = reply.get("frequency")
        try:
            if evaluation_algorithm(risk) == "french":
                self.context.effect = reply.get("severity")
            else:  # Kinney method
                self.context.effect = reply.get("effect")
                self.context.probability = reply.get("probability")
            calculate_priority(self.context, risk)
        except TypeError:
            pass
        return self.context.priority


class ActionPlanView(BrowserView):
    """Logic for creating new action plans.
    """

    phase = "actionplan"
    variation_class = "variation-risk-assessment"
    # The question filter will find modules AND risks
    question_filter = model.ACTION_PLAN_FILTER
    # The risk filter will only find risks
    risk_filter = model.RISK_PRESENT_OR_TOP5_FILTER
    # Skip evaluation?
    # The default value is False, can be overwritten by certain conditions
    skip_evaluation = False
    # Which fields should be skipped? Default are none, i.e. show all
    skip_fields = []
    # What extra style to use for buttons like "Add measure". Default is None.
    style_buttons = None

    @property
    @memoize
    def webhelpers(self):
        return self.context.restrictedTraverse('webhelpers')

    def get_existing_measures(self):
        if not self.use_existing_measures:
            return {}
        if not self.risk or not IRisk.providedBy(self.risk):
            defined_measures = []
        else:
            defined_measures = (
                self.risk.get_pre_defined_measures(self.request) or "")
        try:
            saved_existing_measures = (
                self.context.existing_measures and
                loads(self.context.existing_measures) or [])
            # Backwards compat. We used to save dicts before we
            # switched to list of tuples.
            if isinstance(saved_existing_measures, dict):
                saved_existing_measures = [
                    (k, v) for (k, v) in saved_existing_measures.items()]

            saved_measure_texts = OrderedDict()
            for text, on in saved_existing_measures:
                saved_measure_texts.update({text: on})

            existing_measures = []
            # Pick the pre-defined measures first
            for text in defined_measures:
                active = saved_measure_texts.get(text)
                if active is not None:
                    # Only add the measures that are active
                    if active:
                        existing_measures.append((text, 1))
                    saved_measure_texts.pop(text)

            # Finally, add the user-defined measures as well
            for text, active in saved_measure_texts.items():
                if active:
                    existing_measures.append((text, active))
        except ValueError:
            existing_measures = [
                (text, 0) for text in defined_measures
            ]
            self.context.existing_measures = safe_unicode(
                dumps(existing_measures))
        return existing_measures

    @property
    def risk_present(self):
        return self.context.identification == "no"

    @property
    def is_custom_risk(self):
        return getattr(self.context, 'is_custom_risk', False)

    @property
    def use_problem_description(self):
        if self.is_custom_risk:
            return False
        if IItalyActionPlanPhaseSkinLayer.providedBy(self.request):
            return False
        risk = self.request.survey.restrictedTraverse(
            self.context.zodb_path.split("/"))
        text = risk.problem_description
        return bool(text and text.strip())

    def _extractViewData(self):
        """Extract the data from the current context and build a data structure
        that is usable by the view.
        """

    def _fieldsToDate(self, year, month, day):
        if not day or not year:
            return None
        return datetime.date(year, month, day)

    def __call__(self):
        # Render the page only if the user has edit rights,
        # otherwise redirect to the start page of the session.
        if not (
            self.webhelpers.can_edit_session()
        ):
            return self.request.response.redirect(
                self.webhelpers.survey_url() + '/@@start'
            )
        if redirectOnSurveyUpdate(self.request):
            return
        context = aq_inner(self.context)

        if self.context.is_custom_risk:
            self.risk = self.context
        else:
            self.risk = self.request.survey.restrictedTraverse(
                context.zodb_path.split("/"))

        appconfig = getUtility(IAppConfig)
        settings = appconfig.get('euphorie')
        self.tti = getUtility(IToolTypesInfo)
        self.my_tool_type = get_tool_type(self.context)
        self.use_existing_measures = (
            asBool(settings.get('use_existing_measures', False)) and
            self.my_tool_type in self.tti.types_existing_measures
        )

        self.next_is_report = self.previous_is_identification = False
        # already compute "next" here, so that we can know in the template
        # if the next step might be the report phase, in which case we
        # need to switch off the sidebar
        next = FindNextQuestion(
            context, filter=self.risk_filter)
        if next is None:
            # We ran out of questions, proceed to the report
            url = "%s/report" % self.request.survey.absolute_url()
            self.next_is_report = True
        else:
            url = QuestionURL(
                self.request.survey, next, phase="actionplan")

        previous = FindPreviousQuestion(
            context, filter=self.risk_filter)
        if previous is None:
            # We ran out of questions, step back to identification phase
            previous_url = "%s/identification" % self.request.survey.absolute_url()
            self.previous_is_identification = True
        else:
            previous_url = QuestionURL(
                self.request.survey, previous, phase="actionplan")

        if self.request.environ["REQUEST_METHOD"] == "POST":
            reply = self.request.form
            session = Session()
            context.comment = reply.get("comment")
            context.priority = reply.get("priority")

            new_plans, changes = self.extract_plans_from_request()
            for plan in context.action_plans:
                session.delete(plan)
            context.action_plans.extend(new_plans)
            if changes:
                SessionManager.session.touch()

            if reply["next"] == "previous":
                url = previous_url
            return self.request.response.redirect(url)

        else:
            self.data = context
            if len(context.action_plans) == 0:
                self.data.empty_action_plan = [model.ActionPlan()]

        self.title = context.parent.title
        self.tree = getTreeData(
            self.request, context,
            filter=self.question_filter, phase="actionplan")

        # Italian special
        if IItalyActionPlanPhaseSkinLayer.providedBy(self.request):
            self.skip_evaluation = True
            measures_full_text = True
        else:
            measures_full_text = False
        if self.is_custom_risk:
            self.risk.description = u""
            number_images = 0
        else:
            number_images = getattr(self.risk, 'image', None) and 1 or 0
            if number_images:
                for i in range(2, 5):
                    number_images += getattr(
                        self.risk, 'image{0}'.format(i), None) and 1 or 0
            existing_measures = [
                txt.strip() for (txt, active) in self.get_existing_measures()
                if active]
            solutions = []
            for solution in self.risk.values():
                if not ISolution.providedBy(solution):
                    continue
                if measures_full_text:
                    match = u"%s: %s" % (
                        (getattr(solution, "description", "") or "").strip(),
                        (getattr(solution, "prevention_plan", "") or "").strip()
                    )
                else:
                    match = solution.description.strip()
                if match not in existing_measures:
                    solutions.append(
                        {
                            "description": StripMarkup(solution.description),
                            "action_plan": solution.action_plan,
                            "prevention_plan": solution.prevention_plan,
                            "requirements": solution.requirements,
                        }
                    )
            self.solutions = solutions

        self.number_images = number_images
        self.has_images = number_images > 0
        self.image_class = IMAGE_CLASS[number_images]
        self.risk_number = self.context.number
        lang = getattr(self.request, 'LANGUAGE', 'en')
        if "-" in lang:
            elems = lang.split("-")
            lang = "{0}_{1}".format(elems[0], elems[1].upper())
        self.delete_confirmation = translate(_(
            u"Are you sure you want to delete this measure? This action can "
            u"not be reverted."),
            target_language=lang)
        self.override_confirmation = translate(_(
            u"The current text in the fields 'Action plan', 'Prevention plan' and "
            u"'Requirements' of this measure will be overwritten. This action cannot be "
            u"reverted. Are you sure you want to continue?"),
            target_language=lang)
        self.message_date_before = translate(_(
            u"error_validation_before_end_date",
            default=u"This date must be on or before the end date."),
            target_language=lang)
        self.message_date_after = translate(_(
            u"error_validation_after_start_date",
            default=u"This date must be on or after the start date."),
            target_language=lang)
        self.message_positive_number = translate(_(
            u"error_validation_positive_whole_number",
            default=u"This value must be a positive whole number."),
            target_language=lang)
        return self.index()

    def extract_plans_from_request(self):
        """ Create new ActionPlan objects by parsing the Request.
        """
        new_plans = []
        added = 0
        updated = 0
        existing_plans = {}
        for plan in self.context.action_plans:
            existing_plans[str(plan.id)] = plan
        form = self.request.form
        form["action_plans"] = []
        for i in range(0, len(form.get('measure', []))):
            measure = dict([p for p in form['measure'][i].items()
                            if p[1].strip()])
            form['action_plans'].append(measure)
            if len(measure):
                budget = measure.get("budget")
                budget = budget and budget.split(',')[0].split('.')[0]
                p_start = measure.get('planning_start')
                if p_start:
                    try:
                        datetime.datetime.strptime(p_start, '%Y-%m-%d')
                    except ValueError:
                        p_start = None
                p_end = measure.get('planning_end')
                if p_end:
                    try:
                        datetime.datetime.strptime(p_end, '%Y-%m-%d')
                    except ValueError:
                        p_end = None
                if measure.get('id', '-1') in existing_plans:
                    plan = existing_plans[measure.get('id')]
                    if (
                        measure.get("action_plan") != plan.action_plan or
                        measure.get("prevention_plan") != plan.prevention_plan or  # noqa
                        measure.get("requirements") != plan.requirements or
                        measure.get("responsible") != plan .responsible or (
                            plan.budget and (budget != str(plan.budget)) or
                            plan.budget is None and budget
                        ) or (
                            (plan.planning_start and
                                p_start != plan.planning_start.strftime('%Y-%m-%d')) or  # noqa
                            (plan.planning_start is None and p_start)
                        ) or (
                            (plan.planning_end and
                                p_end != plan.planning_end.strftime('%Y-%m-%d')) or  # noqa
                            (plan.planning_end is None and p_end)
                        )
                    ):
                        updated += 1
                    del existing_plans[measure.get('id')]
                else:
                    added += 1
                new_plans.append(
                    model.ActionPlan(
                        action_plan=measure.get("action_plan"),
                        prevention_plan=measure.get("prevention_plan"),
                        requirements=measure.get("requirements"),
                        responsible=measure.get("responsible"),
                        budget=budget,
                        planning_start=p_start,
                        planning_end=p_end,
                    )
                )
        removed = len(existing_plans)
        changes = True
        if added == 0 and updated == 0 and removed == 0:
            IStatusMessage(self.request).add(
                _(u"No changes were made to measures in your action plan."),
                type='info'
            )
            changes = False
        if added == 1:
            IStatusMessage(self.request).add(
                _(u"message_measure_saved", default=u"A measure has been added to your action plan."),
                type='success'
            )
        elif added == 2:
            IStatusMessage(self.request).add(
                _(
                    u"message_measures_saved_2",
                    default=u"${no_of_measures} measures have been added to your action plan.",
                    mapping={'no_of_measures': str(added)}),
                type='success'
            )
        elif added in (3, 4):
            IStatusMessage(self.request).add(
                _(
                    u"message_measures_saved_3_4",
                    default=u"${no_of_measures} measures have been added to your action plan.",
                    mapping={'no_of_measures': str(added)}),
                type='success'
            )
        elif added > 4:
            IStatusMessage(self.request).add(
                _(
                    u"message_measures_saved",
                    default=u"${no_of_measures} measures have been added to your action plan.",
                    mapping={'no_of_measures': str(added)}),
                type='success'
            )

        if updated == 1:
            IStatusMessage(self.request).add(
                _(u"message_measure_updated", default=u"A measure has been updated in your action plan."),
                type='success'
            )
        elif updated == 2:
            IStatusMessage(self.request).add(
                _(
                    u"message_measures_updated_2",
                    default=u"${no_of_measures} measures have been updated in your action plan.",
                    mapping={'no_of_measures': str(updated)}),
                type='success'
            )
        elif updated in (3, 4):
            IStatusMessage(self.request).add(
                _(
                    u"message_measures_updated_3_4",
                    default=u"${no_of_measures} measures have been updated in your action plan.",
                    mapping={'no_of_measures': str(updated)}),
                type='success'
            )
        elif updated > 4:
            IStatusMessage(self.request).add(
                _(
                    u"message_measures_updated",
                    default=u"${no_of_measures} measures have been updated in your action plan.",
                    mapping={'no_of_measures': str(updated)}),
                type='success'
            )

        if removed == 1:
            IStatusMessage(self.request).add(
                _(u"message_measure_removed", default=u"A measure has been removed from your action plan."),
                type='success'
            )
        elif removed == 2:
            IStatusMessage(self.request).add(
                _(
                    u"message_measures_removed_3",
                    default=u"${no_of_measures} measures have been removed from your action plan.",
                    mapping={'no_of_measures': str(removed)}),
                type='success'
            )
        elif removed in (3, 4):
            IStatusMessage(self.request).add(
                _(
                    u"message_measures_removed_3_4",
                    default=u"${no_of_measures} measures have been removed from your action plan.",
                    mapping={'no_of_measures': str(removed)}),
                type='success'
            )
        elif removed > 4:
            IStatusMessage(self.request).add(
                _(
                    u"message_measures_removed",
                    default=u"${no_of_measures} measures have been removed from your action plan.",
                    mapping={'no_of_measures': str(removed)}),
                type='success'
            )
        return (new_plans, changes)


def calculate_priority(db_risk, risk):
    """Update the risk priority.

    This method can be used for risks using a calculated evaluation method
    to determine the priority absed on the subquestions.
    """
    assert risk.evaluation_method == 'calculated'
    if risk.type in ['top5', 'policy']:
        db_risk.priority = 'high'
    elif evaluation_algorithm(risk) == 'french':
        priority = db_risk.frequency * db_risk.effect
        if priority < 10:
            db_risk.priority = 'low'
        elif priority <= 45:
            db_risk.priority = 'medium'
        else:
            db_risk.priority = 'high'
    else:
        priority = db_risk.frequency * db_risk.effect * db_risk.probability
        if priority <= 15:
            db_risk.priority = 'low'
        elif priority <= 50:
            db_risk.priority = 'medium'
        else:
            db_risk.priority = 'high'
    return db_risk.priority


def evaluation_algorithm(risk):
    for parent in aq_chain(aq_inner(risk)):
        if ISurvey.providedBy(parent):
            return getattr(parent, 'evaluation_algorithm', u'kinney')
    else:
        return u'kinney'


class ConfirmationDeleteRisk(BrowserView):
    """View name: @@confirmation-delete-risk
    """

    @property
    def risk_title(self):
        return self.context.title

    @property
    def risk_id(self):
        return self.context.id

    @property
    def form_action(self):
        return "{0}/@@delete-risk".format(
            aq_parent(self.context).absolute_url())

    def __call__(self, *args, **kwargs):
        ''' Before rendering check if we can find session title
        '''
        self.risk_title
        self.no_splash = True
        return super(ConfirmationDeleteRisk, self).__call__(*args, **kwargs)


class DeleteRisk(BrowserView):
    """View name: @@delete-session
    """

    def __call__(self):
        risk_id = self.request.form.get('risk_id', None)
        if risk_id:
            try:
                risk_id = int(risk_id)
            except ValueError:
                pass
            else:
                keep_ids = [
                    risk.id for risk in self.context.children().all()
                    if risk.id != risk_id]
                self.context.removeChildren(excluded=keep_ids)

        self.request.response.redirect(self.context.absolute_url())



# coding=utf-8
from .. import context_menu

import mock
import unittest


class context_menu_tests(unittest.TestCase):

    def context_menu(self, *a, **kw):
        return context_menu(*a, **kw)

    def test_empty_menu(self):
        with mock.patch(
            'euphorie.client.api.getTreeData', return_value={
                'children': []
            }
        ):
            self.assertEqual(
                self.context_menu(mock.Mock(), 'context', 'phase', None), []
            )

    def test_risk_status_no_class_set(self):
        menu = {
            'children': [{
                'id': 15,
                'type': 'risk',
                'class': None,
                'leaf_module': True,
                'path': '/1',
                'current_parent': False,
                'url': '/1',
                'children': [],
            }]
        }
        request = mock.Mock()
        request.survey_session.absolute_url.return_value = 'http://localhost'
        with mock.patch('euphorie.client.api.getTreeData', return_value=menu):
            context_menu = self.context_menu(request, 'context', 'phase', None)
            self.assertEqual(context_menu[0]['status'], None)

    def test_risk_status_postponed(self):
        menu = {
            'children': [{
                'id': 15,
                'type': 'risk',
                'class': 'postponed',
                'leaf_module': True,
                'path': '/1',
                'current_parent': False,
                'url': '/1',
                'children': [],
            }]
        }
        request = mock.Mock()
        request.survey_session.absolute_url.return_value = 'http://localhost'
        with mock.patch('euphorie.client.api.getTreeData', return_value=menu):
            context_menu = self.context_menu(request, 'context', 'phase', None)
            self.assertEqual(context_menu[0]['status'], 'postponed')

    def test_risk_status_risk_present(self):
        menu = {
            'children': [{
                'id': 15,
                'type': 'risk',
                'class': 'answered risk',
                'leaf_module': True,
                'path': '/1',
                'current_parent': False,
                'url': '/1',
                'children': [],
            }]
        }
        request = mock.Mock()
        request.survey_session.absolute_url.return_value = 'http://localhost'
        with mock.patch('euphorie.client.api.getTreeData', return_value=menu):
            context_menu = self.context_menu(request, 'context', 'phase', None)
            self.assertEqual(context_menu[0]['status'], 'present')

    def test_risk_status_risk_not_present(self):
        menu = {
            'children': [{
                'id': 15,
                'type': 'risk',
                'class': 'current answered',
                'leaf_module': True,
                'path': '/1',
                'current_parent': False,
                'url': '/1',
                'children': [],
            }]
        }
        request = mock.Mock()
        request.survey_session.absolute_url.return_value = 'http://localhost'
        with mock.patch('euphorie.client.api.getTreeData', return_value=menu):
            context_menu = self.context_menu(request, 'context', 'phase', None)
            self.assertEqual(context_menu[0]['status'], 'not-present')

    def test_risk_status_risk_seen(self):
        menu = {
            'children': [{
                'id': 15,
                'type': 'risk',
                'class': 'current',
                'leaf_module': True,
                'path': '/1',
                'current_parent': False,
                'url': '/1',
                'children': [],
            }]
        }
        request = mock.Mock()
        request.survey_session.absolute_url.return_value = 'http://localhost'
        with mock.patch('euphorie.client.api.getTreeData', return_value=menu):
            context_menu = self.context_menu(request, 'context', 'phase', None)
            self.assertEqual(context_menu[0]['status'], None)



from euphorie.client.model import Risk
from euphorie.client.browser.risk import IdentificationView
from euphorie.content.risk import IFrenchEvaluation as IFr
from euphorie.content.risk import IKinneyEvaluation as IKi
from euphorie.content.survey import Survey

import mock
import unittest


class EvaluationViewTests(unittest.TestCase):

    def EvaluationView(self, **kw):
        """
        Note: the EvaluationView does not exist as a separate view any more.
        Its functionality was merged to the identification view.
        For these tests, we just keep the name EvaluationView but return
        IdentificationView, since it contains all the relevant code
        """
        return IdentificationView(Risk(**kw), None)

    def Risk(self, **kw):
        return Risk(**kw)

    def reply(self, freq, effect, prob=None):
        if prob is None:  # French
            return {
                "frequency": IFr['default_frequency'
                                 ].vocabulary.getTermByToken(freq).value,
                "severity": IFr['default_severity']
                .vocabulary.getTermByToken(effect).value,
            }
        else:  # Kinney
            return {
                "frequency": IKi['default_frequency'
                                 ].vocabulary.getTermByToken(freq).value,
                "effect": IKi['default_effect'
                              ].vocabulary.getTermByToken(effect).value,
                "probability": IKi['default_probability']
                .vocabulary.getTermByToken(prob).value,
            }

    def test_evaluation_algorithm_fallback(self):
        view = self.EvaluationView()
        risk = self.Risk()
        self.assertEqual(view.evaluation_algorithm(risk), u"kinney")

    def test_evaluation_algorithm_survey_parent(self):
        view = self.EvaluationView()
        survey = Survey()
        survey.evaluation_algorithm = u"dummy"
        risk = self.Risk().__of__(survey)
        self.assertEqual(view.evaluation_algorithm(risk), u"dummy")

    def test_calculatePriority_kinney_nothing_set(self):
        view = self.EvaluationView()
        risk = self.Risk()
        risk.evaluation_method = 'calculated'
        with mock.patch(
            'euphorie.client.browser.risk.evaluation_algorithm',
            return_value='kinney'
        ):
            self.assertEqual(view.calculatePriority(risk, {}), None)

    def test_calculatePriority_french_nothing_set(self):
        view = self.EvaluationView()
        risk = self.Risk()
        risk.evaluation_method = 'calculated'
        with mock.patch(
            'euphorie.client.browser.risk.evaluation_algorithm',
            return_value='french'
        ):
            self.assertEqual(view.calculatePriority(risk, {}), None)

    def test_calculatePriority_french(self):
        view = self.EvaluationView()
        with mock.patch(
            'euphorie.client.browser.risk.evaluation_algorithm',
            return_value='french'
        ):
            risk = self.Risk()
            risk.evaluation_method = 'calculated'
            # Risks with weak severity are always low priority
            for freq in ["rare", "not-often", "often", "regularly"]:
                self.assertEqual(
                    view.calculatePriority(risk, self.reply(freq, "weak")),
                    "low"
                )
            # High priority items
            self.assertEqual(
                view.calculatePriority(risk, self.reply("often", "severe")),
                "high"
            )
            self.assertEqual(
                view.calculatePriority(
                    risk, self.reply("often", "very-severe")
                ), "high"
            )
            self.assertEqual(
                view.calculatePriority(
                    risk, self.reply("regularly", "severe")
                ), "high"
            )
            self.assertEqual(
                view.calculatePriority(
                    risk, self.reply("regularly", "very-severe")
                ), "high"
            )
            # Some medium priority items
            self.assertEqual(
                view.calculatePriority(
                    risk, self.reply("rare", "very-severe")
                ), "medium"
            )
            self.assertEqual(
                view.calculatePriority(
                    risk, self.reply("not-often", "not-severe")
                ), "medium"
            )

    def test_calculatePriority_kinney(self):
        view = self.EvaluationView()
        risk = self.Risk()
        risk.evaluation_method = 'calculated'
        with mock.patch(
            'euphorie.client.browser.risk.evaluation_algorithm',
            return_value='kinney'
        ):
            # Risks with weak severity are always low priority
            for freq in ["almost-never", "regular", "constant"]:
                self.assertEqual(
                    view.calculatePriority(
                        risk, self.reply(freq, "weak", 'small')
                    ), "low"
                )
            self.assertEqual(
                view.calculatePriority(
                    risk, self.reply("constant", "significant", 'medium')
                ), "high"
            )
            self.assertEqual(
                view.calculatePriority(
                    risk, self.reply("constant", "high", 'medium')
                ), "high"
            )
            self.assertEqual(
                view.calculatePriority(
                    risk, self.reply("constant", "weak", 'medium')
                ), "medium"
            )



# coding=utf-8
from AccessControl.PermissionRole import _what_not_even_god_should_do
from euphorie.client import config
from euphorie.client import model
from euphorie.client.tests.database import DatabaseTests
from euphorie.testing import EuphorieIntegrationTestCase
from sqlalchemy.exc import StatementError
from z3c.saconfig import Session


def createSurvey():
    session = Session()
    account = model.Account(loginname=u"jane", password=u"secret")
    session.add(account)
    survey = model.SurveySession(
        title=u"Session", zodb_path="survey", account=account
    )
    session.add(survey)
    session.flush()
    return (session, survey)


class SurveySessionTests(EuphorieIntegrationTestCase):

    def test_iface(self):
        ''' SurveySessions are marked by the ISurveySession interface
        '''
        survey = createSurvey()[-1]
        self.assertTrue(model.ISurveySession.providedBy(survey))

    def testNoChildren(self):
        (ses, survey) = createSurvey()
        root = survey.addChild(
            model.Module(title=u"Root", module_id="1", zodb_path="1")
        )
        ses.add(root)
        ses.flush()
        self.assertEqual(root.children().count(), 0)

    def testAddChild(self):
        (ses, survey) = createSurvey()
        root = survey.addChild(
            model.Module(title=u"Root", module_id="1", zodb_path="1")
        )
        ses.add(root)
        root.addChild(
            model.Module(title=u"Module", module_id="1", zodb_path="1/1")
        )
        ses.flush()
        self.assertEqual(root.children().count(), 1)

    def testChildOrder(self):
        (ses, survey) = createSurvey()
        root = survey.addChild(
            model.Module(title=u"Root", module_id="1", zodb_path="1")
        )
        ses.add(root)
        ses.flush()
        root.addChild(
            model.Module(title=u"Profile 5", module_id="5", zodb_path="1/5")
        )
        root.addChild(
            model.Module(title=u"Profile 1", module_id="1", zodb_path="1/1")
        )
        root.addChild(
            model.Module(title=u"Profile 3", module_id="3", zodb_path="1/3")
        )
        ses.flush()
        self.assertEqual([c.module_id for c in list(root.children())],
                         [u"5", "1", "3"])

    def testReset_NoChildren(self):
        (ses, survey) = createSurvey()
        survey.reset()
        children = ses.query(model.SurveyTreeItem.id
                             ).filter(model.SurveyTreeItem.session == survey)
        self.assertEqual(children.count(), 0)

    def testReset_SingleChild(self):
        (ses, survey) = createSurvey()
        root = survey.addChild(
            model.Module(title=u"Root", module_id="1", zodb_path="1")
        )
        ses.add(root)
        children = ses.query(model.SurveyTreeItem.id
                             ).filter(model.SurveyTreeItem.session == survey)
        self.assertEqual(children.count(), 1)
        survey.reset()
        self.assertEqual(children.count(), 0)

    def testHasTree_NoChildren(self):
        (ses, survey) = createSurvey()
        self.assertEqual(survey.hasTree(), False)

    def testHasTree_SingleChild(self):
        (ses, survey) = createSurvey()
        root = survey.addChild(
            model.Module(title=u"Root", module_id="1", zodb_path="1")
        )
        ses.add(root)
        self.assertEqual(survey.hasTree(), True)


class RiskPresentFilterTests(EuphorieIntegrationTestCase):

    def createData(self):
        (self.session, self.survey) = createSurvey()
        self.mod1 = model.Module(
            title=u"Module 1",
            module_id="1",
            zodb_path="1",
            skip_children=False
        )
        self.survey.addChild(self.mod1)
        self.q1 = model.Risk(
            title=u"Risk 1",
            risk_id="1",
            zodb_path="1/1",
            type="risk",
            identification="no"
        )
        self.mod1.addChild(self.q1)

    def query(self):
        return self.session.query(model.SurveyTreeItem).filter(
            model.RISK_PRESENT_FILTER
        )

    def testValidRisk(self):
        self.createData()
        self.assertEqual(self.query().count(), 1)

    def testIncludeTop5(self):
        self.createData()
        self.q1.risk_type = "top5"
        self.assertEqual(self.query().count(), 1)

    def testNoPresentRisk(self):
        self.createData()
        self.q1.identification = "n/a"
        self.assertEqual(self.query().count(), 0)


class RiskPresentNoTop5FilterTests(EuphorieIntegrationTestCase):

    def createData(self):
        (self.session, self.survey) = createSurvey()
        self.mod1 = model.Module(
            title=u"Module 1",
            module_id="1",
            zodb_path="1",
            skip_children=False
        )
        self.survey.addChild(self.mod1)
        self.q1 = model.Risk(
            title=u"Risk 1",
            risk_id="1",
            zodb_path="1/1",
            type="risk",
            identification="no"
        )
        self.mod1.addChild(self.q1)

    def query(self):
        return self.session.query(model.SurveyTreeItem)\
                .filter(model.RISK_PRESENT_NO_TOP5_NO_POLICY_DO_EVALUTE_FILTER)

    def testValidRisk(self):
        self.createData()
        self.assertEqual(self.query().count(), 1)

    def testSkipTop5(self):
        self.createData()
        self.q1.risk_type = "top5"
        self.assertEqual(self.query().count(), 0)

    def testSkipPolicy(self):
        self.createData()
        self.q1.risk_type = "policy"
        self.assertEqual(self.query().count(), 0)

    def testNoPresentRisk(self):
        self.createData()
        self.q1.identification = "n/a"
        self.assertEqual(self.query().count(), 0)


class ModuleWithRiskFilterTests(EuphorieIntegrationTestCase):

    def createData(self):
        (self.session, self.survey) = createSurvey()
        self.mod1 = model.Module(
            title=u"Module 1",
            module_id="1",
            zodb_path="1",
            skip_children=False
        )
        self.survey.addChild(self.mod1)
        self.q1 = model.Risk(
            title=u"Risk 1",
            risk_id="1",
            zodb_path="1/1",
            type="risk",
            identification="no"
        )
        self.mod1.addChild(self.q1)

    def query(self):
        return self.session.query(model.SurveyTreeItem)\
                .filter(model.MODULE_WITH_RISK_FILTER)

    def testValidModule(self):
        self.createData()
        self.assertEqual(self.query().count(), 1)

    def testSkipChildren(self):
        self.createData()
        self.mod1.skip_children = True
        self.assertEqual(self.query().count(), 0)

    def testNoPresentRisk(self):
        self.createData()
        self.q1.identification = "n/a"
        self.assertEqual(self.query().count(), 0)

    def testNoChildren(self):
        (self.session, self.survey) = createSurvey()
        self.session.add(self.survey)
        self.mod1 = model.Module(
            title=u"Module 1",
            module_id="1",
            zodb_path="1",
            skip_children=False
        )
        self.survey.addChild(self.mod1)
        self.assertEqual(self.query().count(), 0)


class AccountTests(DatabaseTests):

    def testGlobalRoles(self):
        user = model.Account()
        self.assertEqual(user.getRoles(), ("EuphorieUser", ))

    def testRolesInContext(self):
        user = model.Account()
        self.assertEqual(user.getRolesInContext(None), ("EuphorieUser", ))

    def testNoViewPermission(self):
        user = model.Account()
        self.assertEqual(user.has_permission("View", None), False)

    def testHasSurveyViewPermission(self):
        user = model.Account()
        self.assertEqual(
            user.has_permission("Euphorie: View a Survey", None), True
        )

    def testNonLicetLoviNonLicetBovi(self):
        user = model.Account()
        self.assertEqual(
            user.allowed(None, _what_not_even_god_should_do), False
        )

    def testHasAnonymousRole(self):
        user = model.Account()
        self.assertEqual(user.allowed(None, "Anonymous"), True)

    def testHasAuthenticatedRole(self):
        user = model.Account()
        self.assertEqual(user.allowed(None, "Authenticated"), True)

    def testHasEuphorieUserRole(self):
        user = model.Account()
        self.assertEqual(user.allowed(None, "EuphorieUser"), True)

    def testNoOtherRole(self):
        user = model.Account()
        self.assertEqual(user.allowed(None, "Manager"), False)

    def testAccountType(self):
        (self.session, self.survey) = createSurvey()
        account = self.survey.account
        self.assertEqual(account.account_type, None)
        account.account_type = config.GUEST_ACCOUNT
        self.assertEqual(account.account_type, config.GUEST_ACCOUNT)
        self.session.flush()  # check that exception is not raised
        account.account_type = config.CONVERTED_ACCOUNT
        self.session.flush()  # check that exception is not raised
        self.assertEqual(account.account_type, config.CONVERTED_ACCOUNT)
        account.account_type = 'invalid'
        self.assertRaises(StatementError, self.session.flush)

    def testGroups(self):
        session = Session()
        group1 = model.Group(group_id='1')
        session.add(group1)
        session.flush()
        self.assertEqual(group1.group_id, '1')
        # Verify that a group might have one parent but many children
        group2 = model.Group(group_id='2')
        session.add(group2)
        group3 = model.Group(group_id='3')
        session.add(group3)
        group3.parent = group2.parent = group1
        session.flush()
        self.assertEqual(group2.group_id, '2')
        self.assertEqual(group2.parent, group1)
        self.assertEqual(group1.parent, None)
        self.assertListEqual(group1.children, [group2, group3])
        self.assertListEqual(group2.children, [])

    def testAccountGroupsRelationship(self):
        session = Session()
        group1 = model.Group(group_id='1')
        session.add(group1)
        session.flush()
        account1 = model.Account(loginname='account1')
        session.add(account1)
        account2 = model.Account(loginname='account2')
        session.add(account2)
        session.flush()
        self.assertEqual(account1.group, None)
        account1.group = group1
        account2.group = group1
        self.assertListEqual(group1.accounts, [account1, account2])

    def testAccountGroupsHierarchy(self):
        session = Session()
        group1 = model.Group(group_id='1')
        session.add(group1)
        group2 = model.Group(group_id='2')
        group2.parent = group1
        session.add(group2)
        group3 = model.Group(group_id='3')
        session.add(group3)
        group3.parent = group1
        group4 = model.Group(group_id='4')
        session.add(group3)
        group4.parent = group2
        session.flush()
        account1 = model.Account(loginname='account1')
        session.add(account1)
        account1.group = group1
        account2 = model.Account(loginname='account2')
        session.add(account2)
        account2.group = group2
        account3 = model.Account(loginname='account3')
        session.add(account3)
        account3.group = group3
        account4 = model.Account(loginname='account4')
        session.add(account4)
        session.flush()
        self.assertListEqual(group1.descendants, [group2, group3, group4])
        self.assertListEqual(group2.descendants, [group4])
        self.assertListEqual(group3.descendants, [])
        self.assertListEqual(account1.groups, [group1, group2, group3, group4])
        self.assertListEqual(account2.groups, [group2, group4])
        self.assertListEqual(account3.groups, [group3])
        self.assertListEqual(account4.groups, [])
        self.assertListEqual(group1.parents, [])
        self.assertListEqual(group2.parents, [group1])
        self.assertListEqual(group3.parents, [group1])
        self.assertListEqual(group4.parents, [group2, group1])

    def testSessions(self):
        session = Session()
        group1 = model.Group(group_id='1')
        session.add(group1)
        group2 = model.Group(group_id='2')
        session.add(group2)
        account1 = model.Account(loginname='account1')
        session.add(account1)
        account1.group = group1
        group2.parent = group1
        from functools import partial
        add_survey = partial(model.SurveySession, account=account1)
        survey1 = add_survey(zodb_path='1')
        session.add(survey1)
        survey2 = add_survey(zodb_path='2', group=group1)
        session.add(survey2)
        survey3 = add_survey(zodb_path='3', group=group2)
        session.add(survey3)
        session.flush()
        self.assertListEqual(account1.sessions, [survey1, survey2, survey3])
        self.assertListEqual(group1.sessions, [survey2])
        self.assertListEqual(group2.sessions, [survey3])

    def testSessionAcquisition(self):
        ''' Users belonging to a group should be able to see all the sessions
        belonging to the group and the group children
        '''
        session = Session()
        group1 = model.Group(group_id='1')
        session.add(group1)
        group2 = model.Group(group_id='2')
        session.add(group2)
        account1 = model.Account(loginname='account1')
        session.add(account1)
        account1.group = group1
        group2.parent = group1
        account2 = model.Account(loginname='account2')
        session.add(account2)
        account2.group = group2
        survey1 = model.SurveySession(
            account=account1,
            group=group1,
            zodb_path='1',
        )
        session.add(survey1)
        survey2 = model.SurveySession(
            account=account2,
            group=group2,
            zodb_path='2',
        )
        session.add(survey2)
        session.flush()
        self.assertListEqual(account1.sessions, [survey1])
        self.assertListEqual(account2.sessions, [survey2])
        self.assertListEqual(account1.acquired_sessions, [survey1, survey2])
        self.assertListEqual(account2.acquired_sessions, [survey2])



# coding=utf-8
from euphorie.client.api.authentication import authenticate_token
from euphorie.client.api.authentication import generate_token
from euphorie.client.model import Account
from euphorie.client.tests.database import DatabaseTests
from euphorie.client.tests.utils import addAccount
from plone.keyring.interfaces import IKeyManager

import mock
import unittest


class generate_token_tests(unittest.TestCase):

    def generate_token(self, *a, **kw):
        return generate_token(*a, **kw)

    def test_token_depends_on_login(self):
        with mock.patch('euphorie.client.api.authentication.getUtility') \
                as mock_getUtility:
            mock_getUtility(IKeyManager).secret.return_value = 'secret'
            self.assertNotEqual(
                self.generate_token(Account(id=5, loginname='john')),
                self.generate_token(Account(id=5, loginname='jane'))
            )

    def test_token_depends_on_password(self):
        with mock.patch('euphorie.client.api.authentication.getUtility') \
                as mock_getUtility:
            mock_getUtility(IKeyManager).secret.return_value = 'secret'
            self.assertNotEqual(
                self.generate_token(
                    Account(id=5, loginname='john', password='jane')
                ),
                self.generate_token(
                    Account(id=5, loginname='john', password='alice')
                )
            )

    def test_password_not_in_token(self):
        with mock.patch('euphorie.client.api.authentication.getUtility') \
                as mock_getUtility:
            mock_getUtility(IKeyManager).secret.return_value = 'secret'
            self.assertTrue(
                'jane' not in self.generate_token(
                    Account(id=5, loginname='john', password='jane')
                )
            )


class authenticate_token_tests(DatabaseTests):

    def authenticate_token(self, *a, **kw):
        return authenticate_token(*a, **kw)

    def test_bad_token_form(self):
        self.assertEqual(self.authenticate_token('a-b-c'), None)

    def test_bad_id_in_token(self):
        self.assertEqual(self.authenticate_token('a-bc'), None)

    def test_unknown_account(self):
        self.assertEqual(self.authenticate_token('5-bc'), None)

    def test_known_account(self):
        with mock.patch(
            'euphorie.client.api.authentication.generate_token',
            return_value='1-hash'
        ):
            addAccount()
            self.assertEqual(
                self.authenticate_token('1-hash'), ('1', 'jane@example.com')
            )

    def test_invalid_token(self):
        with mock.patch(
            'euphorie.client.api.authentication.generate_token',
            return_value='1-otherhash'
        ):
            addAccount()
            self.assertTrue(self.authenticate_token('1-hash') is None)



# coding=utf-8
from cStringIO import StringIO
from euphorie.client import model
from euphorie.client.model import Account
from euphorie.client.model import SurveySession
from euphorie.client.report import ActionPlanTimeline
from euphorie.client.report import HtmlToRtf
from euphorie.client.report import IdentificationReport
from euphorie.client.tests.utils import testRequest
from euphorie.client.utils import setRequest
from euphorie.content.risk import Risk
from euphorie.testing import EuphorieIntegrationTestCase
from rtfng.document.section import Section
from rtfng.Elements import Document
from rtfng.Renderer import Renderer
from z3c.saconfig import Session

import datetime
import mock
import unittest


class IdentificationReportTests(unittest.TestCase):

    def IdentificationReport(self, *a, **kw):
        return IdentificationReport(*a, **kw)

    def test_title_not_a_risk(self):
        node = mock.Mock()
        node.type = 'module'
        node.title = u'My title'
        view = self.IdentificationReport(None, None)
        self.assertEqual(view.title(node, None), u'My title')

    def test_title_unanswered_risk(self):
        node = mock.Mock()
        node.type = 'risk'
        node.identification = None
        node.title = u'My title'
        view = self.IdentificationReport(None, None)
        self.assertEqual(view.title(node, None), u'My title')

    def test_title_empty_problem_description(self):
        node = mock.Mock()
        node.type = 'risk'
        node.identification = u'no'
        node.title = u'My title'
        zodb_node = mock.Mock()
        zodb_node.problem_description = u'   '
        view = self.IdentificationReport(None, None)
        self.assertEqual(view.title(node, zodb_node), u'My title')

    def test_title_risk_present_and_with_problem_description(self):
        node = mock.Mock()
        node.type = 'risk'
        node.identification = u'no'
        node.title = u'My title'
        zodb_node = mock.Mock()
        zodb_node.problem_description = u'Bad situation'
        view = self.IdentificationReport(None, None)
        self.assertEqual(view.title(node, zodb_node), u'Bad situation')


class ShowNegateWarningTests(unittest.TestCase):

    def _call(self, node, zodbnode):
        report = IdentificationReport(None, None)
        return report.show_negate_warning(node, zodbnode)

    def test_show_Unanswered(self):
        # https//code.simplon.biz/tracker/tno-euphorie/ticket/75
        zodbnode = Risk()
        zodbnode.problem_description = None
        node = model.Risk(type="risk")
        self.assertEqual(self._call(node, zodbnode), False)

    def test_RiskNotPresent(self):
        zodbnode = Risk()
        zodbnode.problem_description = None
        node = model.Risk(type="risk", identification="yes")
        self.assertEqual(self._call(node, zodbnode), False)

    def test_RiskNotApplicable(self):
        zodbnode = Risk()
        zodbnode.problem_description = None
        node = model.Risk(type="risk", identification="n/a")
        self.assertEqual(self._call(node, zodbnode), False)

    def test_Present(self):
        zodbnode = Risk()
        zodbnode.problem_description = None
        node = model.Risk(type="risk", identification="no")
        self.assertEqual(self._call(node, zodbnode), True)

    def test_HasProblemDescription(self):
        zodbnode = Risk()
        zodbnode.problem_description = u"Negative"
        node = model.Risk(type="risk", identification="no")
        self.assertEqual(self._call(node, zodbnode), False)

    def test_HasEmptyProblemDescription(self):
        zodbnode = Risk()
        zodbnode.problem_description = u"   "
        node = model.Risk(type="risk", identification="no")
        self.assertEqual(self._call(node, zodbnode), True)


class HtmlToRtfTests(unittest.TestCase):

    def HtmlToRtf(self, *a, **kw):
        return HtmlToRtf(*a, **kw)

    def render(self, output):
        document = Document()
        section = Section()
        for o in output:
            section.append(o)
        document.Sections.append(section)
        renderer = Renderer()
        renderer.Write(document, StringIO())  # Setup instance variables
        renderer._doc = document
        renderer._fout = StringIO()
        renderer._CurrentStyle = ""
        renderer._WriteSection(section, True, False)
        return renderer._fout.getvalue()

    def testEmptyInput(self):
        self.assertEqual(self.HtmlToRtf(u"", u"<stylesheet>"), [])

    def testInvalidHtmlFallback(self):
        self.assertTrue(
            "text\\par" in
            self.render(self.HtmlToRtf(u"<p>text</p>", u"<stylesheet>"))
        )

    def testBasicParagraph(self):
        self.assertTrue(
            "Simple text\\par" in self.render(
                self.HtmlToRtf(u"<p>Simple text</p>", u"<stylesheet>")
            ), []
        )

    def testItalicInText(self):
        self.assertTrue(
            "Simple {\\i text}\\par" in self.render(
                self.
                HtmlToRtf(u"<p>Simple <em>text</em></p>", u"<stylesheet>")
            )
        )

    def testBoldAndItalicText(self):
        self.assertTrue(
            "Very {\\i very }{\\b\\i bold}\\par" in self.render(
                self.HtmlToRtf(
                    u"<p>Very <em>very <strong>bold</strong></em></p>",
                    u"<stylesheet>"
                )
            )
        )

    def testEmphasisInText(self):
        self.assertTrue(
            "{\\i text}" in
            self.render(self.HtmlToRtf(u"<em>text</em>", u"<stylesheet>"))
        )

    def testInlineEntity(self):
        self.assertTrue(
            "Simple & clean\\par" in self.render(
                self.HtmlToRtf(u"<p>Simple &amp; clean</p>", u"<stylesheet>")
            )
        )

    def testInlineEntityDigit(self):
        self.assertTrue(
            "Simple \r clean\\par" in self.render(
                self.HtmlToRtf(u"<p>Simple &#13; clean</p>", u"<stylesheet>")
            )
        )

    def test_link_in_text(self):
        # This demonstrates TNO Euphorie ticket 186
        html = '<p>Check the <a rel="nofollow">manual</a> for more info.</p>'
        rendering = self.render(self.HtmlToRtf(html, '<stylesheet>'))
        self.assertTrue('Check the manual for more info.' in rendering)
        self.assertEqual(rendering.count('more info'), 1)


class ActionPlanTimelineTests(EuphorieIntegrationTestCase):

    def ActionPlanTimeline(self, *a, **kw):
        return ActionPlanTimeline(*a, **kw)

    def _create_session(self, dbsession, loginname='jane'):
        session = SurveySession(
            account=Account(loginname=loginname, password=u'john'),
            zodb_path='survey'
        )
        dbsession.add(session)
        return session

    def test_get_measures_with_correct_module(self):
        dbsession = Session()
        session = self._create_session(dbsession)
        # This first module should be ignored, it doesn't contain any risks
        session.addChild(model.Module(
            zodb_path='1',
            module_id='1',
        ))
        # Between the next two modules, the first one (root-level) must be
        # returned.
        module = session.addChild(
            model.Module(
                zodb_path='2',
                module_id='2',
            )
        )
        module = module.addChild(
            model.Module(
                zodb_path='2/3',
                module_id='3',
            )
        )
        module.addChild(
            model.Risk(
                zodb_path='2/3/4',
                risk_id='1',
                identification='no'
            )
        )
        request = testRequest()
        request.survey = mock.Mock()
        request.survey.restrictedTraverse = lambda x: object
        request.survey.ProfileQuestions = lambda: []
        view = self.ActionPlanTimeline(None, request)
        view.session = session
        measures = view.get_measures()
        self.assertEqual(len(measures), 1)
        self.assertEqual(measures[0][0].module_id, u'2')

    def test_get_measures_return_risks_without_measures(self):
        dbsession = Session()
        session = self._create_session(dbsession)
        module = session.addChild(
            model.Module(
                session=session,
                zodb_path='1',
                module_id='1',
            )
        )
        module.addChild(
            model.Risk(
                session=session,
                zodb_path='1/2',
                risk_id='1',
                identification='no'
            )
        )
        request = testRequest()
        request.survey = mock.Mock()
        request.survey.restrictedTraverse = lambda x: object
        request.survey.ProfileQuestions = lambda: []
        setRequest(request)
        view = self.ActionPlanTimeline(None, request)
        view.session = session
        measures = view.get_measures()
        self.assertEqual(len(measures), 1)
        self.assertEqual(measures[0][2], None)

    def test_get_measures_filter_on_session(self):
        dbsession = Session()
        sessions = []
        for login in ['jane', 'john']:
            session = self._create_session(dbsession, loginname=login)
            module = session.addChild(
                model.Module(
                    session=session,
                    zodb_path='1',
                    module_id='1',
                )
            )
            module.addChild(
                model.Risk(
                    session=session,
                    zodb_path='1/2',
                    risk_id='1',
                    identification='no',
                    action_plans=[
                        model.ActionPlan(
                            action_plan=u'Measure 1 for %s' % login)
                    ]))
            sessions.append(session)

        request = testRequest()
        request.survey = mock.Mock()
        request.survey.restrictedTraverse = lambda x: object
        request.survey.ProfileQuestions = lambda: []
        setRequest(request)
        view = self.ActionPlanTimeline(None, request)
        view.session = sessions[0]
        measures = view.get_measures()
        self.assertEqual(len(measures), 1)
        self.assertEqual(measures[0][2].action_plan, 'Measure 1 for jane')

    def test_get_measures_order_by_start_date(self):
        dbsession = Session()
        session = self._create_session(dbsession)
        module = session.addChild(
            model.Module(
                session=session,
                zodb_path='1',
                module_id='1',
            )
        )
        module.addChild(
            model.Risk(
                session=session,
                zodb_path='1/2',
                risk_id='1',
                identification='no',
                action_plans=[
                    model.ActionPlan(
                        action_plan=u'Plan 2',
                        planning_start=datetime.date(2011, 12, 15)
                    ),
                    model.ActionPlan(
                        action_plan=u'Plan 1',
                        planning_start=datetime.date(2011, 11, 15)
                    )
                ]
            )
        )

        request = testRequest()
        request.survey = mock.Mock()
        request.survey.restrictedTraverse = lambda x: object
        request.survey.ProfileQuestions = lambda: []
        setRequest(request)
        view = self.ActionPlanTimeline(None, request)
        view.session = session
        measures = view.get_measures()
        self.assertEqual(len(measures), 2)
        self.assertEqual([row[2].action_plan for row in measures],
                         [u'Plan 1', u'Plan 2'])

    def test_priority_name_known_priority(self):
        view = self.ActionPlanTimeline(None, None)
        self.assertEqual(view.priority_name('high'), u'High')

    def test_priority_name_known_unpriority(self):
        view = self.ActionPlanTimeline(None, None)
        self.assertEqual(view.priority_name('dummy'), 'dummy')

    def test_create_workbook_empty_session(self):
        # If there are no risks only the header row should be generated.
        request = testRequest()
        request.survey = None
        setRequest(request)
        view = self.ActionPlanTimeline(None, request)
        view.getModulePaths = lambda: []
        book = view.create_workbook()
        self.assertEqual(len(book.worksheets), 1)
        sheet = book.worksheets[0]
        self.assertEqual(len(sheet.rows), 1)

    def test_create_workbook_plan_information(self):
        dbsession = Session()
        session = self._create_session(dbsession)
        module = model.Module(
            zodb_path='1',
            title=u'Top-level Module title',
        )
        risk = model.Risk(
            zodb_path='1/2/3',
            risk_id='1',
            title=u'Risk title',
            priority='high',
            identification='no',
            path='001002003',
            comment=u'Risk comment'
        )
        plan = model.ActionPlan(
            action_plan=u'Plan 2',
            planning_start=datetime.date(2011, 12, 15),
            budget=500
        )
        request = testRequest()
        request.survey = mock.Mock()
        zodb_node = mock.Mock()
        zodb_node.problem_description = u'This is wrong.'
        request.survey.restrictedTraverse.return_value = zodb_node
        view = self.ActionPlanTimeline(None, request)
        view.session = session
        view.get_measures = lambda: [(module, risk, plan)]
        wb = view.create_workbook()
        sheet = wb.worksheets[0]
        # planning start
        self.assertEqual(
            sheet.cell('A2').value.date(), datetime.date(2011, 12, 15)
        )
        # planning end
        self.assertEqual(sheet.cell('B2').value, None)
        # action plan
        self.assertEqual(sheet.cell('C2').value, u'Plan 2')
        # prevention plan
        self.assertEqual(sheet.cell('D2').value, None)
        # requirements
        self.assertEqual(sheet.cell('E2').value, None)
        # responsible
        self.assertEqual(sheet.cell('F2').value, None)
        # budget
        self.assertEqual(sheet.cell('G2').value, 500)
        # module title
        self.assertEqual(sheet.cell('H2').value, u'Top-level Module title')
        # risk number
        self.assertEqual(sheet.cell('I2').value, u'1.2.3')
        # risk title
        self.assertEqual(sheet.cell('J2').value, u'This is wrong.')
        # risk priority
        self.assertEqual(sheet.cell('K2').value, u'High')
        # risk comment
        self.assertEqual(sheet.cell('L2').value, u'Risk comment')

    def test_create_workbook_no_problem_description(self):
        dbsession = Session()
        session = self._create_session(dbsession)
        module = model.Module(
            zodb_path='1',
            path='001',
            title=u'Top-level Module title',
        )
        risk = model.Risk(
            zodb_path='1/2/3',
            risk_id='1',
            title=u'Risk title',
            priority='high',
            identification='no',
            path='001002003',
            comment=u'Risk comment')
        request = testRequest()
        request.survey = mock.Mock()
        request.survey.ProfileQuestions = lambda: []
        zodb_node = mock.Mock()
        zodb_node.title = u'Risk title.'
        zodb_node.problem_description = u'  '
        request.survey.restrictedTraverse.return_value = zodb_node
        setRequest(request)
        view = self.ActionPlanTimeline(None, request)
        view.session = session
        view.getRisks = lambda x: [(module, risk)]
        sheet = view.create_workbook().worksheets[0]
        self.assertEqual(sheet.cell('J2').value, u'Risk title')

    def test_render_value(self):
        request = testRequest()
        request.survey = None
        view = self.ActionPlanTimeline(None, request)
        view.session = SurveySession(title=u'Acme')
        view.render()
        response = request.response
        self.assertEqual(
            response.headers['content-type'], 'application/vnd.openxmlformats-'
            'officedocument.spreadsheetml.sheet'
        )
        self.assertEqual(
            response.headers['content-disposition'],
            'attachment; filename="Timeline for Acme.xlsx"'
        )



"""
Survey views
------------
"""
from .. import MessageFactory as _
from ..ghost import PathGhost
from Acquisition import aq_chain
from Acquisition import aq_inner
from Acquisition import aq_parent
from collections import defaultdict
from datetime import datetime
from decimal import Decimal
from euphorie.client import model
from euphorie.client import utils
from euphorie.client.country import IClientCountry
from euphorie.client.interfaces import IActionPlanPhaseSkinLayer
from euphorie.client.interfaces import IClientSkinLayer
from euphorie.client.interfaces import ICustomizationPhaseSkinLayer
from euphorie.client.interfaces import IEvaluationPhaseSkinLayer
from euphorie.client.interfaces import IFranceActionPlanPhaseSkinLayer
from euphorie.client.interfaces import IFranceCustomizationPhaseSkinLayer
from euphorie.client.interfaces import IFranceEvaluationPhaseSkinLayer
from euphorie.client.interfaces import IFranceIdentificationPhaseSkinLayer
from euphorie.client.interfaces import IFranceReportPhaseSkinLayer
from euphorie.client.interfaces import IIdentificationPhaseSkinLayer
from euphorie.client.interfaces import IItalyActionPlanPhaseSkinLayer
from euphorie.client.interfaces import IItalyCustomizationPhaseSkinLayer
from euphorie.client.interfaces import IItalyEvaluationPhaseSkinLayer
from euphorie.client.interfaces import IItalyIdentificationPhaseSkinLayer
from euphorie.client.interfaces import IItalyReportPhaseSkinLayer
from euphorie.client.interfaces import IReportPhaseSkinLayer
from euphorie.client.navigation import FindFirstQuestion
from euphorie.client.navigation import getTreeData
from euphorie.client.navigation import QuestionURL
from euphorie.client.profile import extractProfile
from euphorie.client.session import SessionManager
from euphorie.client.update import redirectOnSurveyUpdate
from euphorie.content.interfaces import ICustomRisksModule
from euphorie.content.survey import ISurvey
from five import grok
from plone import api
from plone.memoize.instance import memoize
from plone.memoize.view import memoize_contextless
from sqlalchemy import case
from sqlalchemy import func
from sqlalchemy import orm
from sqlalchemy import sql
from z3c.appconfig.interfaces import IAppConfig
from z3c.saconfig import Session
from zope.component import adapts
from zope.component import getUtility
from zope.i18n import translate
from zope.i18nmessageid import MessageFactory
from zope.interface import directlyProvidedBy
from zope.interface import directlyProvides
from ZPublisher.BaseRequest import DefaultPublishTraverse

import decimal
import logging


PloneLocalesFactory = MessageFactory("plonelocales")

log = logging.getLogger(__name__)

grok.templatedir("templates")


class Resume(grok.View):
    """Survey resume screen.

    This view is used when a user resumes an existing session.

    View name: @@resume
    """
    grok.context(ISurvey)
    grok.require("euphorie.client.ViewSurvey")
    grok.layer(IClientSkinLayer)
    grok.name("resume")

    def render(self):
        """ We always open the tool on the "Start" page.
            Formerly, we would jump to the first question. Since the Start
            page is now much more important, it is also used when a session
            gets resumed.
        """
        survey = aq_inner(self.context)
        if redirectOnSurveyUpdate(self.request):
            return

        self.request.response.redirect(
            "%s/start?initial_view=1" % survey.absolute_url()
        )


class Identification(grok.View):
    """Survey identification start page.

    This view shows the introduction text for the identification phase. This
    includes an option to print a report with all questions.

    This view is registered for :py:class:`PathGhost` instead of
    :py:obj:`euphorie.content.survey.ISurvey` since the
    :py:class:`SurveyPublishTraverser` generates a :py:class:`PathGhost` object
    for the *identification* component of the URL.

    View name: @@index_html
    """
    grok.context(PathGhost)
    grok.require("euphorie.client.ViewSurvey")
    grok.layer(IIdentificationPhaseSkinLayer)
    grok.template("identification")
    grok.name("index_html")
    variation_class = "variation-risk-assessment"

    question_filter = None

    def update(self):
        self.next_url = None
        if redirectOnSurveyUpdate(self.request):
            return
        self.survey = survey = aq_parent(aq_inner(self.context))
        question = FindFirstQuestion(filter=self.question_filter)
        if question is not None:
            self.next_url = QuestionURL(
                survey, question, phase="identification"
            )
            self.tree = getTreeData(self.request, question)

    @property
    def extra_text(self):
        appconfig = getUtility(IAppConfig)
        settings = appconfig.get('euphorie')
        have_extra = settings.get('extra_text_identification', False)
        if not have_extra:
            return None
        lang = getattr(self.request, 'LANGUAGE', 'en')
        # Special handling for Flemish, for which LANGUAGE is "nl-be". For
        # translating the date under plone locales, we reduce to generic "nl".
        # For the specific oira translation, we rewrite to "nl_BE"
        if "-" in lang:
            elems = lang.split("-")
            lang = "{0}_{1}".format(elems[0], elems[1].upper())
        return translate(
            _(u"extra_text_identification", default=u""), target_language=lang
        )


class ActionPlan(grok.View):
    """Survey action plan start page.

    This view shows the introduction text for the action plan phase.

    This view is registered for :py:class:`PathGhost` instead of
    :py:obj:`euphorie.content.survey.ISurvey` since the
    :py:class:`SurveyPublishTraverser` generates a
    :py:class:`PathGhost` object for the *actionplan* component of the URL.
    """
    grok.context(PathGhost)
    grok.require("euphorie.client.ViewSurvey")
    grok.layer(IActionPlanPhaseSkinLayer)
    grok.template("actionplan")
    grok.name("index_html")
    variation_class = "variation-risk-assessment"

    # The question filter will find modules AND risks
    question_filter = model.ACTION_PLAN_FILTER
    # The risk filter will only find risks
    risk_filter = model.RISK_PRESENT_OR_TOP5_FILTER

    def update(self):
        if redirectOnSurveyUpdate(self.request):
            return
        self.survey = survey = aq_parent(aq_inner(self.context))
        # We fetch the first actual risk, so that we can jump directly to it.
        question = FindFirstQuestion(filter=self.risk_filter)
        if question is not None:
            # We also fetch the first module, so that we can properly build the
            # tree: Open at the first module, but with no risk being selected
            module = FindFirstQuestion(filter=self.question_filter)
            self.next_url = QuestionURL(survey, question, phase="actionplan")
            self.tree = getTreeData(
                self.request,
                module,
                filter=self.question_filter,
                phase="actionplan"
            )
        else:
            self.next_url = None

    def __call__(self):
        ''' Render the page only if the user has edit rights,
        otherwise redirect to the start page of the session.
        '''
        if self.context.restrictedTraverse('webhelpers').can_edit_session():
            return super(ActionPlan, self).__call__()
        return self.request.response.redirect(
            self.context.aq_parent.absolute_url() + '/@@start'
        )


class _StatusHelper(object):

    COUNTRIES_WITHOUT_HIGH_RISKS = [
        'it',
    ]

    @property
    @memoize
    def sql_session(self):
        return Session()

    @property
    @memoize_contextless
    def preferred_language(self):
        return api.portal.get_tool('portal_languages').getPreferredLanguage()

    def module_query(self, sessionid, optional_modules):
        if optional_modules:
            case_clause = case([
                (
                    sql.and_(
                        model.SurveyTreeItem.profile_index != -1,
                        model.SurveyTreeItem.zodb_path.in_(optional_modules)
                    ), func.substr(model.SurveyTreeItem.path, 1, 6)
                ),
                (
                    sql.and_(
                        model.SurveyTreeItem.profile_index == -1,
                        model.SurveyTreeItem.zodb_path.in_(optional_modules)
                    ), func.substr(model.SurveyTreeItem.path, 1, 3) +
                    '000-profile'
                ),
                (
                    sql.and_(
                        model.SurveyTreeItem.profile_index != -1,
                        model.SurveyTreeItem.depth < 2
                    ), func.substr(model.SurveyTreeItem.path, 1, 3)
                ),
            ])
        else:
            case_clause = case([
                (
                    sql.and_(
                        model.SurveyTreeItem.profile_index != -1,
                        model.SurveyTreeItem.depth < 2
                    ), func.substr(model.SurveyTreeItem.path, 1, 3)
                ),
            ])

        query = self.sql_session.query(case_clause.label('module')).filter(
            sql.and_(
                model.SurveyTreeItem.session_id == sessionid,
                model.SurveyTreeItem.type == 'module'
            )
        ).group_by('module').order_by('module')
        return query

    def slicePath(self, path):
        while path:
            yield path[:3].lstrip("0")
            path = path[3:]

    def getModulePaths(self):
        """ Return a list of all the top-level modules belonging to this survey.
        """
        session_id = self.session.id
        if not session_id:
            return []
        profile = extractProfile(self.request.survey, SessionManager.session)
        module_query = self.module_query(
            sessionid=session_id, optional_modules=profile.keys()
        )
        module_res = module_query.all()
        modules_and_profiles = {}
        for row in module_res:
            if row.module is not None:
                if row.module.find('profile') > 0:
                    path = row.module[:3]
                    modules_and_profiles[path] = 'profile'
                else:
                    modules_and_profiles[row[0]] = ''
        module_paths = [m.module for m in module_res if m.module is not None]
        module_paths = modules_and_profiles.keys()
        module_paths = sorted(module_paths)
        self.modules_and_profiles = modules_and_profiles
        return module_paths

    def getModules(self):
        """ Return a list of dicts of all the top-level modules and locations
            belonging to this survey.
        """
        sql_session = self.sql_session
        session_id = self.session.id
        module_paths = self.getModulePaths()
        base_url = "%s/identification" % self.request.survey.absolute_url()
        parent_node = orm.aliased(model.Module)
        titles = dict(
            sql_session.query(
                model.Module.path, model.Module.title
            ).filter(model.Module.session_id == session_id).filter(
                model.Module.path.in_(module_paths)
            )
        )

        location_titles = dict(
            sql_session.query(
                model.Module.path, parent_node.title
            ).filter(model.Module.session_id == session_id).filter(
                model.Module.path.in_(module_paths)
            ).filter(
                sql.and_(
                    parent_node.session_id == session_id,
                    parent_node.depth < model.Module.depth,
                    model.Module.path.like(parent_node.path + "%")
                )
            )
        )
        modules = {}
        toc = {}
        title_custom_risks = utils.get_translated_custom_risks_title(
            self.request
        )

        for path in module_paths:
            number = ".".join(self.slicePath(path))
            # top-level module, always include it in the toc
            if len(path) == 3:
                title = titles[path]
                if title == 'title_other_risks' or title == u"Other risks":
                    title = title_custom_risks
                toc[path] = {
                    'path': path,
                    'title': title,
                    'locations': [],
                    'number': number,
                }
                # If this is a profile (aka container for locations), skip
                # adding to the list of modules
                if self.modules_and_profiles[path] == 'profile':
                    continue
            # sub-module (location) or location container
            else:
                if path in location_titles:
                    title = u"{0} - {1}".format(
                        location_titles[path], titles[path]
                    )
                    toc[path[:3]]['locations'].append({
                        'path': path,
                        'title': titles[path],
                        'number': number,
                    })
                else:
                    log.warning(
                        "Status: found a path for a submodule {0} for which "
                        "there's no location title.".format(path)
                    )
                    continue

            modules[path] = {
                'path': path,
                'title': title,
                'url': '%s/%s' % (base_url, '/'.join(self.slicePath(path))),
                'todo': 0,
                'ok': 0,
                'postponed': 0,
                'risk_with_measures': 0,
                'risk_without_measures': 0,
                'number': number,
            }
        self.tocdata = toc
        return modules

    def getRisks(self, module_paths, skip_unanswered=False):
        """ Return a list of risk dicts for risks that belong to the modules
            with paths as specified in module_paths.
        """
        global request
        request = self.request
        if not len(module_paths):
            return []
        sql_session = self.sql_session
        session_id = self.session.id
        # First, we need to compute the actual module paths, making sure that
        # skipped optional modules are excluded
        # This means top-level module paths like 001 or 001002 can be replaced
        # by several sub-modules paths like 001002, 001003 and 001002001
        path_clause = [
            model.SurveyTreeItem.path.like('{}%'.format(mp))
            for mp in module_paths
        ]

        module_query = sql_session.query(model.SurveyTreeItem).filter(
            sql.and_(
                model.SurveyTreeItem.session_id == session_id,
                model.SurveyTreeItem.type == 'module', sql.or_(*path_clause)
            )
        ).order_by(model.SurveyTreeItem.path)

        module_res = module_query.all()
        modules_by_path = {m.path: m for m in module_res}

        def nodes(modules):
            global use_nodes, s_paths
            use_nodes = []
            s_paths = sorted(modules, key=lambda x: x.path)
            # In case of repeatable profile questions, the top-level module
            # path will be 6 digits long.
            top_nodes = [
                elem for elem in s_paths if (
                    len(elem.path) == 3 or
                    (len(elem.path) == 6 and elem.path[:3] not in s_paths)
                ) and not elem.skip_children
            ]

            def use_node(elem):
                # Recursively find the nodes that are not disabled
                global use_nodes
                # Skip this elem?
                # If this is an optional module, check the "postponed" flag.
                # As long as the optional question has not been answered, skip
                # showing its children.
                # Only a "Yes" answer on the module will be considered as "do
                # not skip children"
                zodb_elem = request.survey.restrictedTraverse(
                    elem.zodb_path.split('/')
                )
                if getattr(zodb_elem, 'optional', False):
                    if (
                        (
                            elem.postponed in (True, None) or
                            elem.skip_children
                        ) and not ICustomRisksModule.providedBy(zodb_elem)
                    ):
                        return
                children = [
                    x for x in s_paths if x.path.startswith(elem.path) and
                    len(x.path) == len(elem.path) + 3
                ]
                if children:
                    for child in children:
                        use_node(child)
                else:
                    use_nodes.append(elem.path)
            for elem in top_nodes:
                use_node(elem)
            ret = []
            # Here we make sure that only the longest paths of sub-modules
            # are used, but not the parents. Example
            # (001, 002, 001001, 001003) will be turned into
            # (001001, 001003, 002), since the parent 001 contains sub-modules,
            # and some of those might have been de-selected, like 001002
            for elem in sorted(use_nodes, reverse=True):
                if not [x for x in ret if x.startswith(elem)]:
                    ret.append(elem)
            return ret

        filtered_module_paths = nodes(tuple(module_res))

        child_node = orm.aliased(model.Risk)
        risks = sql_session.query(model.Module, model.Risk).filter(
            sql.and_(
                model.Module.session_id == session_id,
                model.Module.path.in_(filtered_module_paths),
                sql.and_(
                    child_node.session_id == model.Module.session_id,
                    child_node.depth > model.Module.depth,
                    child_node.path.like(model.Module.path + "%")
                )
            )
        ).join((
            model.Risk,
            sql.and_(
                model.Risk.path.startswith(model.Module.path),
                model.Risk.session_id == session_id
            )
        )).order_by(model.Risk.path)

        def _module_path(path):
            # Due to the extended query above that replaces top-module paths
            # with sub-module paths (if present), we need to cut back the path
            # under which we store each risk back to the original top-level
            # module path
            for mp in module_paths:
                if path.startswith(mp):
                    return mp

        filtered_risks = []
        for (module, risk) in risks.all():
            if risk.identification != 'n/a':
                if (
                    skip_unanswered and
                    risk.identification is None and
                    not risk.postponed
                ):
                    continue
                module_path = _module_path(module.path)
                # And, since we might have truncated the path to represent
                # the top-level module, we also need to get the corresponding
                # module object.
                module = modules_by_path[module_path]
                filtered_risks.append((module, risk))
        return filtered_risks

    def as_decimal(self, num):
        return decimal.Decimal(num)

    @property
    @memoize
    def show_high_risks(self):
        for obj in aq_chain(aq_inner(self.context)):
            if IClientCountry.providedBy(obj):
                if obj.id in self.COUNTRIES_WITHOUT_HIGH_RISKS:
                    return False
        return True


class Status(grok.View, _StatusHelper):
    """Show survey status information.
    """
    grok.context(ISurvey)
    grok.require("euphorie.client.ViewSurvey")
    grok.layer(IClientSkinLayer)
    grok.template("status")
    variation_class = "variation-risk-assessment"

    def __init__(self, context, request):
        super(Status, self).__init__(context, request)

        def default_risks_by_status():
            return {
                'present': {
                    'high': [],
                    'medium': [],
                    'low': [],
                },
                'possible': {
                    'postponed': [],
                    'todo': [],
                },
            }

        self.risks_by_status = defaultdict(default_risks_by_status)
        now = datetime.now()
        lang = date_lang = getattr(self.request, 'LANGUAGE', 'en')
        # Special handling for Flemish, for which LANGUAGE is "nl-be". For
        # translating the date under plone locales, we reduce to generic "nl".
        # For the specific oira translation, we rewrite to "nl_BE"
        if "-" in lang:
            date_lang = lang.split("-")[0]
            elems = lang.split("-")
            lang = "{0}_{1}".format(elems[0], elems[1].upper())
        self.date = u"{0} {1} {2}".format(
            now.strftime('%d'),
            translate(
                PloneLocalesFactory(
                    "month_{0}".format(now.strftime('%b').lower()),
                    default=now.strftime('%B'),
                ),
                target_language=date_lang,
            ), now.strftime('%Y')
        )
        self.label_page = translate(
            _(u"label_page", default=u"Page"), target_language=lang
        )
        self.label_page_of = translate(
            _(u"label_page_of", default=u"of"), target_language=lang
        )
        self.session = SessionManager.session
        if (
            self.session is not None and self.session.title != (
                callable(getattr(self.context, 'Title', None))
                and self.context.Title() or ''
            )
        ):
            self.session_title = self.session.title
        else:
            self.session_title = None

    def getStatus(self):
        """ Gather a list of the modules and locations in this survey as well
            as data around their state of completion.
        """
        session = Session()
        self.session = SessionManager.session
        total_ok = 0
        total_with_measures = 0
        modules = self.getModules()
        filtered_risks = self.getRisks([m['path'] for m in modules.values()])
        for (module, risk) in filtered_risks:
            module_path = module.path
            has_measures = False
            if risk.identification in ['yes', 'n/a']:
                total_ok += 1
                modules[module_path]['ok'] += 1
            elif risk.identification == 'no':
                measures = session.query(
                    model.ActionPlan.id
                ).filter(model.ActionPlan.risk_id == risk.id)
                if measures.count():
                    has_measures = True
                    modules[module_path]['risk_with_measures'] += 1
                    total_with_measures += 1
                else:
                    modules[module_path]['risk_without_measures'] += 1
            elif risk.postponed:
                modules[module_path]['postponed'] += 1
            else:
                modules[module_path]['todo'] += 1

            self.add_to_risk_list(risk, module_path, has_measures=has_measures)

        for key, m in modules.items():
            if m['ok'] + m['postponed'] + m['risk_with_measures'] + m[
                'risk_without_measures'
            ] + m['todo'] == 0:
                del modules[key]
                del self.tocdata[key]
        self.percentage_ok = (
            not len(filtered_risks) and 100 or
            int(
                (total_ok + total_with_measures) /
                Decimal(len(filtered_risks)) * 100
            )
        )
        self.status = modules.values()
        self.status.sort(key=lambda m: m["path"])
        self.toc = self.tocdata.values()
        self.toc.sort(key=lambda m: m["path"])

    def add_to_risk_list(self, risk, module_path, has_measures=False):
        if self.is_skipped_from_risk_list(risk):
            return

        risk_title = self.get_risk_title(risk)

        base_url = "%s/actionplan" % self.request.survey.absolute_url()
        url = '%s/%s' % (base_url, '/'.join(self.slicePath(risk.path)))

        if risk.identification != 'no':
            status = risk.postponed and 'postponed' or 'todo'
            self.risks_by_status[module_path]['possible'][status].append({
                'title': risk_title,
                'path': url
            })
        else:
            self.risks_by_status[module_path
                                 ]['present'][risk.priority or 'low'].append({
                                     'title': risk_title,
                                     'path': url,
                                     'has_measures': has_measures
                                 })

    def get_risk_title(self, risk):
        if risk.is_custom_risk:
            risk_title = risk.title
        else:
            risk_obj = self.request.survey.restrictedTraverse(
                risk.zodb_path.split('/')
            )
            if not risk_obj:
                return
            if risk.identification == 'no':
                risk_title = risk_obj.problem_description
            else:
                risk_title = risk.title
        return risk_title

    def is_skipped_from_risk_list(self, risk):
        if risk.priority == "high":
            if risk.identification != 'no':
                if risk.risk_type not in ['top5']:
                    return True
        else:
            return True

    def update(self):
        if redirectOnSurveyUpdate(self.request):
            return
        self.getStatus()


def find_sql_context(session_id, zodb_path):
    """Find the closest SQL tree node for a candidate path.

    The path has to be given as a list of path entries. The session
    timestamp is only used as part of a cache key for this method.

    The return value is the id of the SQL tree node. All consumed
    entries will be removed from the zodb_path list.
    """
    # Pop all integer elements from the URL
    path = ""
    head = []
    while zodb_path:
        next = zodb_path.pop()
        if len(next) > 3:
            zodb_path.append(next)
            break

        try:
            path += '%03d' % int(next)
            head.append(next)
        except ValueError:
            zodb_path.append(next)
            break

    # Try and find a SQL tree node that matches our URL
    query = (
        Session.query(
            model.SurveyTreeItem.id
        ).filter(model.SurveyTreeItem.session_id == session_id
                 ).filter(model.SurveyTreeItem.path == sql.bindparam('path'))
    )
    while path:
        node = query.params(path=path).first()
        if node is not None:
            return node[0]
        path = path[:-3]
        zodb_path.append(head.pop())


def build_tree_aq_chain(root, tree_id):
    """Build an acquisition context for a tree node.
    """
    tail = Session.query(model.SurveyTreeItem).get(tree_id)
    walker = root
    path = tail.path
    while len(path) > 3:
        id = str(int(path[:3]))
        path = path[3:]
        walker = PathGhost(id).__of__(walker)
    return tail.__of__(walker)


class SurveyPublishTraverser(DefaultPublishTraverse):
    """Publish traverser to setup the survey skin layers.

    This traverser marks the request with IClientSkinLayer. We can not use
    BeforeTraverseEvent sine in Zope 2 that is only fired for site objects.
    """
    adapts(ISurvey, IClientSkinLayer)

    phases = {
        'identification': IIdentificationPhaseSkinLayer,
        'customization': ICustomizationPhaseSkinLayer,
        'evaluation': IEvaluationPhaseSkinLayer,
        'actionplan': IActionPlanPhaseSkinLayer,
        'report': IReportPhaseSkinLayer,
    }

    countries = {
        'it': {
            'identification': IItalyIdentificationPhaseSkinLayer,
            'customization': IItalyCustomizationPhaseSkinLayer,
            'evaluation': IItalyEvaluationPhaseSkinLayer,
            'actionplan': IItalyActionPlanPhaseSkinLayer,
            'report': IItalyReportPhaseSkinLayer,
        },
        'fr': {
            'identification': IFranceIdentificationPhaseSkinLayer,
            'customization': IFranceCustomizationPhaseSkinLayer,
            'evaluation': IFranceEvaluationPhaseSkinLayer,
            'actionplan': IFranceActionPlanPhaseSkinLayer,
            'report': IFranceReportPhaseSkinLayer,
        },
    }

    def hasValidSession(self, request):
        """Check if the user has an active session for the survey.
        """
        dbsession = SessionManager.session
        client_path = utils.RelativePath(request.client, self.context)

        if dbsession is None or \
                dbsession.zodb_path != client_path:

            # Allow for alternative session ids to be hardcoded in the
            # euphorie.ini file for automatic browser testing with Browsera
            conf = getUtility(IAppConfig).get("euphorie", {})
            debug_ids = conf.get('debug_sessions', '').strip().splitlines()
            for sid in debug_ids:
                session = Session.query(SessionManager.model).get(sid)
                if hasattr(session, 'zodb_path') and \
                        session.zodb_path == client_path:
                    SessionManager.resume(session)
                    return True

            return False
        return True

    def publishTraverse(self, request, name):
        request.survey = self.context
        utils.setLanguage(request, self.context, self.context.language)

        if name not in ["view", "index_html"] and \
                not self.hasValidSession(request):
            request.response.redirect(
                aq_parent(aq_parent(self.context)).absolute_url(), lock=True
            )
            return self.context

        if name not in self.phases:
            return super(SurveyPublishTraverser, self)\
                    .publishTraverse(request, name)

        # Decorate the request with the right skin layer and add to the aq path

        # Some countries need to be marked specially. Check if this needs to be
        # done, and decorate the reques accordingly if yes.
        special = False
        for obj in aq_chain(aq_inner(self.context)):
            if IClientCountry.providedBy(obj):
                if obj.id in self.countries:
                    special = True
                    directlyProvides(
                        request, self.countries[obj.id][name],
                        *directlyProvidedBy(request)
                    )
                    break
        if not special:
            directlyProvides(
                request, self.phases[name], *directlyProvidedBy(request)
            )
        self.context = PathGhost(name).__of__(self.context)

        session = SessionManager.session
        tree_id = find_sql_context(
            session.id, request['TraversalRequestNameStack']
        )
        if tree_id is not None:
            return build_tree_aq_chain(self.context, tree_id)

        # No SQL based traversal possible, return the existing context with the
        # new skin layer applied
        return self.context



"""
Appendix
--------

The @@appendix and @@about views.
"""

import logging
from zope.interface import Interface
from five import grok
from euphorie.client.interfaces import IClientSkinLayer

log = logging.getLogger(__name__)

grok.templatedir("templates")


class About(grok.View):
    """View name: @@about
    """
    grok.context(Interface)
    grok.layer(IClientSkinLayer)
    grok.name("about")
    grok.template("about")



# coding=utf-8
from Acquisition import aq_base
from Acquisition import aq_parent
from euphorie.client.api.actionplans import RiskActionPlans
from euphorie.client.api.actionplans import View
from euphorie.client.api.authentication import generate_token
from euphorie.client.api.tests.utils import _setup_session
from euphorie.client.model import ActionPlan
from euphorie.client.model import Risk
from euphorie.client.model import Session
from euphorie.client.model import SurveySession
from euphorie.client.tests.utils import addAccount
from euphorie.testing import EuphorieFunctionalTestCase
from euphorie.testing import EuphorieIntegrationTestCase
from plone import api
from plone.app.testing.interfaces import SITE_OWNER_NAME

import datetime
import json
import mock


class RiskActionPlansTests(EuphorieIntegrationTestCase):

    def RiskActionPlans(self, *a, **kw):
        return RiskActionPlans(*a, **kw)

    def _make_risk(self, login):
        account = addAccount(login=login)
        risk = Risk(
            risk_id='15',
            path='01',
            zodb_path='/foo/bar/1',
            session=SurveySession(account=account, zodb_path='/foo/bar')
        )
        Session.add(risk)
        return risk

    def test_getitem_id_from_other_risk(self):
        risk = self._make_risk('jane')
        plan = ActionPlan(action_plan=u'This is the plan')
        risk.action_plans.append(plan)
        plans = self.RiskActionPlans('actionplans', None, risk)
        wrapped_plan = plans['1']
        self.assertTrue(aq_base(wrapped_plan) is plan)
        self.assertTrue(aq_base(aq_parent(wrapped_plan)) is plans)
        # I am commenting this line because plans['1'] does the same
        # and does not raise.
        # Additionally this test was masked before by another one with the same
        # name...
        # self.assertRaises(KeyError, plans.__getitem__, '1')

    def test_getitem_bad_key(self):
        plans = self.RiskActionPlans('actionplans', None, None)
        self.assertRaises(KeyError, plans.__getitem__, 'xyz')

    def test_getitem_id_from_other_risk2(self):
        risk1 = self._make_risk('jane')
        risk1.action_plans.append(ActionPlan(action_plan=u'This is the plan'))
        risk2 = self._make_risk('cindy')
        plans = self.RiskActionPlans('actionplans', None, risk2)
        self.assertRaises(KeyError, plans.__getitem__, '1')


class ViewTests(EuphorieIntegrationTestCase):

    def _build_context(self, **kw):
        with api.env.adopt_user(SITE_OWNER_NAME):
            (account, survey, survey_session) = _setup_session(self.portal)
        risk = Session.query(Risk).one()
        return RiskActionPlans('actionplans', None, risk)

    def test_plans_no_plans_defined(self):
        context = self._build_context()
        view = View(context, self.request.clone())
        self.assertEqual(view.plans(), [])

    def test_with_plan(self):
        context = self._build_context()
        plan = ActionPlan(action_plan=u'This is the plan')
        context.risk.action_plans.append(plan)
        with mock.patch('euphorie.client.api.actionplans.plan_info') \
                as mock_plan_info:
            mock_plan_info.return_value = 'plan-info'
            view = View(context, self.request.clone())
            plans = view.plans()
            self.assertEqual(plans, ['plan-info'])
            mock_plan_info.assert_called_once_with(plan)

    def test_do_GET_response(self):
        view = View(None, None)
        view.plans = mock.Mock(return_value='plan-info')
        response = view.do_GET()
        self.assertTrue(isinstance(response, dict))
        self.assertEqual(set(response), set(['action-plans']))
        self.assertEqual(response['action-plans'], 'plan-info')
        view.plans.assert_called_once_with()

    def test_do_POST_success(self):
        context = mock.Mock()
        context.risk.action_plans = []
        view = View(context, self.request.clone())
        with mock.patch(
            'euphorie.client.api.actionplan.View.do_PUT',
            return_value={
                'type': 'actionplan'
            }
        ):
            view.input = 'input'
            response = view.do_POST()
            self.assertEqual(response, {'type': 'actionplan'})
            self.assertEqual(len(context.risk.action_plans), 1)

    def test_do_POST_error(self):
        view = View(None, self.request.clone())
        with mock.patch(
            'euphorie.client.api.actionplan.View.do_PUT',
            return_value={
                'type': 'error'
            }
        ):
            view.input = 'input'
            response = view.do_POST()
            self.assertEqual(response, {'type': 'error'})


class BrowserTests(EuphorieFunctionalTestCase):

    def test_get(self):
        with api.env.adopt_user(SITE_OWNER_NAME):
            (account, survey, survey_session) = _setup_session(self.portal)
            Session.add(survey_session)
            risk = Session.query(Risk).one()
            risk.action_plans.append(
                ActionPlan(
                    action_plan=u'This is the plan',
                    planning_start=datetime.date(2012, 6, 3)
                )
            )
        browser = self.get_browser()
        browser.addHeader('X-Euphorie-Token', generate_token(account))
        browser.open(
            'http://nohost/plone/client/api/users/1/'
            'sessions/1/1/1/actionplans'
        )
        self.assertEqual(browser.headers['Content-Type'], 'application/json')
        response = json.loads(browser.contents)
        self.assertTrue('action-plans' in response)



"""
Login
-----

Register new users, login/logout, create a "Guest user" account and convert
existing guest accounts to normal accounts.
"""

from euphorie.client import MessageFactory as _
from ..conditions import approvedTermsAndConditions
from ..conditions import checkTermsAndConditions
from ..country import IClientCountry
from ..session import SessionManager
from ..utils import setLanguage
from AccessControl import getSecurityManager
from Acquisition import aq_chain
from Acquisition import aq_inner
from Acquisition import aq_parent
from euphorie.client import config
from euphorie.client import model
from euphorie.client.browser.country import SessionsView
from euphorie.content.survey import ISurvey
from plone import api
from plone.memoize.view import memoize
from plone.session.plugins.session import cookie_expiration_date
from plonetheme.nuplone.tiles.analytics import trigger_extra_pageview
from Products.CMFCore.utils import getToolByName
from Products.Five import BrowserView
from Products.statusmessages.interfaces import IStatusMessage
from z3c.appconfig.interfaces import IAppConfig
from z3c.appconfig.utils import asBool
from z3c.saconfig import Session
from zope import component
from zope.i18n import translate

import cgi
import datetime
import logging
import os
import re
import six
import urllib
import urlparse


log = logging.getLogger(__name__)

# I know this is a stupid regular expression, but it Works For Us(tm)
EMAIL_RE = re.compile(
    r'[_a-z0-9+-]+(\.[_a-z0-9+-]+)*@[a-z0-9-]+(\.[a-z0-9-]+)*(\.[a-z]{2,4})',
    re.IGNORECASE)


class Login(BrowserView):
    """View name: @@login"""

    variation_class = "variation-dashboard"

    def setLanguage(self, came_from):
        qs = urlparse.urlparse(came_from)[4]
        params = cgi.parse_qs(qs)
        lang = params.get("language")
        if not lang:
            if IClientCountry.providedBy(self.context):
                lang = getattr(self.context, 'language', None)
                if lang and isinstance(lang, six.string_types):
                    lang = [lang]
        if not lang:
            return
        setLanguage(self.request, self.context, lang=lang[0])

    def login(self, account, remember):
        pas = getToolByName(self.context, "acl_users")
        pas.updateCredentials(
            self.request,
            self.request.RESPONSE,
            account.loginname,
            account.password,
        )
        if remember:
            self.request.RESPONSE.cookies['__ac']['expires'] = cookie_expiration_date(120)  # noqa: E501
            self.request.RESPONSE.cookies['__ac']['max_age'] = 120 * 24 * 60 * 60  # noqa: E501

    def transferGuestSession(self, session_id):
        """ Transfer guest session to an existing user account
        """
        if not session_id:
            return
        account = getSecurityManager().getUser()
        session = Session.query(model.SurveySession).get(session_id)
        session.account_id = account.id
        SessionManager.resume(session)

    def __call__(self):
        context = aq_inner(self.context)
        came_from = self.request.form.get("came_from")
        if came_from:
            if isinstance(came_from, list):
                # If came_from is both in the querystring and the form data
                came_from = came_from[0]
            self.setLanguage(came_from)
        else:
            came_from = aq_parent(context).absolute_url()

        account = getSecurityManager().getUser()
        appconfig = component.getUtility(IAppConfig)
        settings = appconfig.get('euphorie')
        self.allow_guest_accounts = asBool(
            settings.get('allow_guest_accounts', False))

        lang = api.portal.get_current_language()
        self.show_whofor = False if lang in ('fr', ) else True
        self.show_what_to_do = False if lang in ('fr', ) else True
        self.show_how_long = False if lang in ('fr', ) else True
        self.show_why_register = True
        self.show_prepare = False if lang in ('fr', ) else True

        if self.request.environ["REQUEST_METHOD"] == "POST":
            reply = self.request.form
            if reply["next"] == "previous":
                next = aq_parent(aq_inner(context)).absolute_url()
                self.request.RESPONSE.redirect(next)
                return

            if (
                isinstance(account, model.Account) and
                account.getUserName() == reply.get("__ac_name", '').lower()
            ):
                self.transferGuestSession(reply.get('guest_session_id'))
                self.login(account, bool(self.request.form.get('remember')))
                v_url = urlparse.urlsplit(
                    self.request.URL + '/success').path.replace("@@", "")
                trigger_extra_pageview(self.request, v_url)

                if (
                    checkTermsAndConditions() and
                    not approvedTermsAndConditions(account)
                ):
                    self.request.RESPONSE.redirect(
                        "%s/terms-and-conditions?%s" % (
                            context.absolute_url(),
                            urllib.urlencode({"came_from": came_from}),
                        )
                    )
                else:
                    self.request.RESPONSE.redirect(came_from)
                return
            self.error = True

        self.reset_password_request_url = "%s/@@reset_password_request?%s" % (
            context.absolute_url(),
            urllib.urlencode({'came_from': came_from}),
        )
        self.register_url = "%s/@@register?%s" % (
            context.absolute_url(),
            urllib.urlencode({'came_from': came_from}),
        )
        self.tryout_url = "%s/@@tryout?%s" % (
            context.absolute_url(),
            urllib.urlencode({'came_from': came_from}),
        )
        return self.index()


class LoginForm(Login):
    """View name: @@login_form"""


class Tryout(Login):
    """Create a guest account

    View name: @@tryout
    """

    def createGuestAccount(self):
        account = model.Account(
            loginname="guest-%s" % datetime.datetime.now().isoformat(),
            account_type=config.GUEST_ACCOUNT
        )
        Session().add(account)
        return account

    def __call__(self):
        came_from = self.request.form.get("came_from")
        if not came_from:
            return self.request.response.redirect(
                api.portal.get().absolute_url()
            )
        account = self.createGuestAccount()
        self.login(account, False)
        client_url = self.request.client.absolute_url()
        came_from = came_from.replace(client_url, '')
        if came_from.startswith('/'):
            came_from = came_from[1:]
        try:
            survey = self.context.restrictedTraverse(came_from)
        except KeyError:
            survey = None
        if not ISurvey.providedBy(survey):
            return self.request.response.redirect(came_from)
        title = survey.Title()
        SessionManager.start(title=title, survey=survey, account=account)
        survey_url = survey.absolute_url()
        v_url = urlparse.urlsplit(survey_url + '/resume').path
        trigger_extra_pageview(self.request, v_url)
        self.request.response.redirect("%s/start" % survey_url)


class CreateTestSession(SessionsView, Tryout):
    """Create a guest session

    View name: @@new-session-test.html
    """

    def __call__(self):
        appconfig = component.getUtility(IAppConfig)
        settings = appconfig.get('euphorie')
        self.allow_guest_accounts = asBool(
            settings.get('allow_guest_accounts', False))
        context = aq_inner(self.context)
        came_from = self.request.form.get("came_from")
        if came_from:
            if isinstance(came_from, list):
                # If came_from is both in the querystring and the form data
                came_from = came_from[0]
        else:
            came_from = context.absolute_url()
        self.register_url = "%s/@@register?%s" % (
            context.absolute_url(), urllib.urlencode({'came_from': came_from}))
        setLanguage(self.request, self.context)
        if self.request.environ["REQUEST_METHOD"] == "POST":
            if not self.allow_guest_accounts:
                flash = IStatusMessage(self.request).addStatusMessage
                flash(_(u"Starting a test session is not available in this OiRA "
                      u"application."), "error")
                self.request.response.redirect(came_from)
            else:
                reply = self.request.form
                if reply["action"] == "new":
                    account = self.createGuestAccount()
                    self.login(account, False)
                    self._NewSurvey(reply, account)
        self._updateSurveys()
        return self.index()


class Register(BrowserView):
    """Register a new account or convert an existing guest account
    """

    def _tryRegistration(self):
        reply = self.request.form
        loginname = reply.get("email")
        if not loginname:
            self.errors["email"] = _("error_missing_email",
                    default=u"Please enter your email address")
        elif not EMAIL_RE.match(loginname):
            self.errors["email"] = _("error_invalid_email",
                    default=u"Please enter a valid email address")
        if not reply.get("password1"):
            self.errors["password"] = _("error_missing_password",
                    default=u"Please enter a password")
        elif reply.get("password1") != reply.get("password2"):
            self.errors["password"] = _("error_password_mismatch",
                    default=u"Passwords do not match")
        if self.errors:
            return False

        session = Session()
        loginname = loginname.lower()
        account = session.query(model.Account)\
                .filter(model.Account.loginname == loginname).count()
        if account:
            self.errors["email"] = _("error_email_in_use",
                default=u"An account with this email address already exists.")
            return False

        pm = getToolByName(self.context, "portal_membership")
        if pm.getMemberById(loginname) is not None:
            self.errors["email"] = _("error_email_in_use",
                default=u"An account with this email address already exists.")
            return False

        guest_session_id = self.request.form.get('guest_session_id')
        if guest_session_id:
            account = getSecurityManager().getUser()
            account.loginname = loginname
            account.password = reply.get("password1")
            account.account_type = config.CONVERTED_ACCOUNT
        else:
            account = model.Account(
                loginname=loginname,
                password=reply.get("password1")
            )
        Session().add(account)
        log.info("Registered new account %s", loginname)
        v_url = urlparse.urlsplit(self.request.URL + '/success').path.replace(
            "@@", "")
        trigger_extra_pageview(self.request, v_url)
        return account

    @property
    @memoize
    def email_message(self):
        lang = getattr(self.request, 'LANGUAGE', 'en')
        if "-" in lang:
            elems = lang.split("-")
            lang = "{0}_{1}".format(elems[0], elems[1].upper())
        return translate(_(
            u"invalid_email",
            default=u"Please enter a valid email address."),
            target_language=lang,
        )

    def __call__(self):
        self.errors = {}
        if self.request.method != "POST":
            return self.index()
        account = self._tryRegistration()
        if not account:
            return self.index()
        pas = getToolByName(self.context, "acl_users")
        pas.updateCredentials(
            self.request,
            self.request.response,
            account.getUserName(),
            account.password,
        )

        country_url = aq_inner(self.context).absolute_url()
        came_from = self.request.form.get("came_from")
        if not came_from:
            came_from = country_url

        if checkTermsAndConditions():
            self.request.response.redirect(
                    "%s/terms-and-conditions?%s" % (
                        self.request.client.absolute_url(),
                        urllib.urlencode({"came_from": came_from})))
        else:
            self.request.response.redirect(came_from)

    def get_image_version(self, name):
        """" Needed on the reports overview show to the guest user """
        fdir = os.path.join(
            os.path.dirname(__file__), os.path.join('templates', 'media'))
        lang = getattr(self.request, 'LANGUAGE', 'en')
        fname = "{0}_{1}".format(name, lang)
        if os.path.isfile(os.path.join(fdir, fname + '.png')):
            return fname
        return name


class Logout(BrowserView):

    def __call__(self):
        SessionManager.stop()

        pas = getToolByName(self.context, "acl_users")
        pas.resetCredentials(self.request, self.request.response)

        for obj in aq_chain(aq_inner(self.context)):
            if IClientCountry.providedBy(obj):
                break
        else:
            obj = self.request.client

        self.request.response.redirect(obj.absolute_url())
